package org.processmining.indatabasepreprocessing.algorithms;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.graphbased.directed.transitionsystem.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.petrinet.behavioralanalysis.CGGenerator;
import org.processmining.plugins.pnml.exporting.PnmlExportNetToPNML;
import org.processmining.plugins.pnml.importing.PnmlImportNet;

public class BehaviouralProfileProduction {
	
	private CoverabilityGraph graph;
	private HashMap<String, Set<String>> eventuallyFollows;
	private HashMap<String, Set<String>> eventuallyFollowsNoDuplication;
	private HashMap<String, Set<String>> exclusiveness;
	private HashMap<String, Set<String>> inverse;
	private Set<String> activities;
	private Set<String> invisibleLabels;
		
	public BehaviouralProfileProduction(UIPluginContext context, Petrinet petrinet, Marking initialMarking) throws Exception {
		this.invisibleLabels = new HashSet<String>();
		this.invisibleLabels.add("$invisible$"); this.invisibleLabels.add("tau start"); this.invisibleLabels.add("tau join"); this.invisibleLabels.add("tau split"); this.invisibleLabels.add("tau from tree");
		
		long startTime = System.currentTimeMillis();
		System.out.println("unfolded the net...");
		Pair<Petrinet, Marking> unfoldedTotal = unfold(context, petrinet);
		
		System.out.println("create coverability graph...");
		graph = context.tryToFindOrConstructFirstNamedObject(
				CoverabilityGraph.class,
				CGGenerator.class.getAnnotation(Plugin.class).name(),
				null,
				null,
				unfoldedTotal.getFirst(),
				unfoldedTotal.getSecond());
		
		// how if the net is not unfolded...
		/*System.out.println("create coverability graph...");
		graph = context.tryToFindOrConstructFirstNamedObject(
				CoverabilityGraph.class,
				CGGenerator.class.getAnnotation(Plugin.class).name(),
				null,
				null,
				petrinet,
				initialMarking);*/
		
		long endTime = System.currentTimeMillis();		
		System.out.println("PREPROCESSING TIME (UNFOLDING & COVERABILITY GRAPH) (in ms): " + (endTime - startTime));
	}
	
	public void createBehaviouralProfile() {
		Set<Transition> edges = graph.getEdges();
		Iterator<Transition> it = edges.iterator();
		activities = new HashSet<String>();
		
		while(it.hasNext()) {
			String label = it.next().getLabel();
			if(!invisibleLabels.contains(label)) {
				// remove copy
				if(label.contains("_copy")) {
					String[] arr = label.split("_copy");
					label = arr[0];
				}
				activities.add(label);
			}
		}
		
		long startTime = System.currentTimeMillis();
		createEventuallyFollows();
		createExclusiveness();
		createInverse();
		long endTime = System.currentTimeMillis();
		
		System.out.println("\nPREPROCESSING TIME (CREATE BEHAVIOURAL PROFILES) (in ms): " + (endTime - startTime));
	}
	
	private void createExclusiveness() {
		exclusiveness = new HashMap<String, Set<String>>();
		
		Iterator<String> itActivities = activities.iterator();
		
		while(itActivities.hasNext()) {
			String x = itActivities.next();
			Set<String> exclusiveSet = new HashSet<String>();
			
			Iterator<String> itOtherActivities = activities.iterator();
			while(itOtherActivities.hasNext()) {
				String y = itOtherActivities.next();
				
				// x + y iff x !> y and y !> x
				if(!eventuallyFollowsNoDuplication.get(x).contains(y) && !eventuallyFollowsNoDuplication.get(y).contains(x)) {
					exclusiveSet.add(y);
				}				
			}
			
			exclusiveness.put(x, exclusiveSet);
		}
	}
	
	private void createInverse() {
		inverse = new HashMap<String, Set<String>>();
		
		Iterator<String> itActivities = activities.iterator();
		
		while(itActivities.hasNext()) {
			String x = itActivities.next();
			Set<String> inverseSet = new HashSet<String>();
			
			Iterator<String> itOtherActivities = activities.iterator();
			while(itOtherActivities.hasNext()) {
				String y = itOtherActivities.next();
				
				// x <- y iff y > x and x !> y
				if(eventuallyFollowsNoDuplication.get(y).contains(x) && !eventuallyFollowsNoDuplication.get(x).contains(y)) {
					inverseSet.add(y);
				}
			}
			
			inverse.put(x, inverseSet);
		}
	}
	
	private void createEventuallyFollows() {
		eventuallyFollows = new HashMap<String, Set<String>>();
		
		Set<State> states = graph.getNodes();
		Iterator<State> itState = states.iterator();
		
		while(itState.hasNext()) {
			State state = itState.next();
			Collection<Transition> transitions = graph.getOutEdges(state);
			Iterator<Transition> itTransition = transitions.iterator();
			
			while(itTransition.hasNext()) {
				Transition transition = itTransition.next();
				String label = transition.getLabel();				
				
				// if not an invisible transition
				if(!invisibleLabels.contains(label)) {
					Queue<State> tobeVisited = new LinkedList<State>();
					Set<State> alreadyVisited = new HashSet<State>();
					tobeVisited.add(transition.getTarget());
					alreadyVisited.add(state);
					Set<String> efTransitions = collectEventuallyFollows(graph, tobeVisited, alreadyVisited, new HashSet<String>());
					
					if(eventuallyFollows.get(transition.getLabel()) != null) {
						Set<String> trans = eventuallyFollows.get(label);
						trans.addAll(efTransitions);
						eventuallyFollows.put(label, trans);
					} else {
						eventuallyFollows.put(label, efTransitions);
					}
				}
				
			}
		}
		
		// remove copy label in eventually follows
		eventuallyFollowsNoDuplication = new HashMap<String, Set<String>>();
		for(String act: eventuallyFollows.keySet()) {
			Set<String> child = eventuallyFollows.get(act);
			Iterator<String> it = child.iterator();
			
			if(act.contains("_copy")) {
				String[] arr = act.split("_copy");
				act = arr[0];
			}
			
			Set<String> childNew = new HashSet<String>();			
			while(it.hasNext()) {
				String act2 = it.next();
				if(act2.contains("_copy")) {
					String[] arr = act2.split("_copy");
					act2 = arr[0];
				}
				childNew.add(act2);
			}
			
			if(eventuallyFollowsNoDuplication.containsKey(act)) {
				eventuallyFollowsNoDuplication.get(act).addAll(childNew);
			} else {
				eventuallyFollowsNoDuplication.put(act, childNew);
			}
			
		}
		
		// print ef		
		for(String key: eventuallyFollowsNoDuplication.keySet()) {
			System.out.print("\n" + key + ":\t");
			Set<String> child = eventuallyFollowsNoDuplication.get(key);
			Iterator<String> it = child.iterator();
			while(it.hasNext()) {
				System.out.print(it.next() + " ");
			}
		}
		
		// print ef		
		System.out.println("WITH DUPLICATION");
		for(String key: eventuallyFollows.keySet()) {
			System.out.print("\n" + key + ":\t");
			Set<String> child = eventuallyFollows.get(key);
			Iterator<String> it = child.iterator();
			while(it.hasNext()) {
				System.out.print(it.next() + " ");
			}
		}
	}
	
	private Set<String> collectEventuallyFollows(CoverabilityGraph graph, Queue<State> tobeVisited, Set<State> alreadyVisited, Set<String> efTransitions) {
		if(tobeVisited.isEmpty()) {
			return efTransitions;
		} else {
			
			State s = tobeVisited.remove();
			//System.out.println("state now: " + s.getLabel());
			if(!alreadyVisited.contains(s)) {
				alreadyVisited.add(s);
				
				if(graph.getOutEdges(s) != null) {
					Collection<Transition> ts = graph.getOutEdges(s);
					Iterator<Transition> it = ts.iterator();
					while(it.hasNext()) {
						Transition t = it.next();
						String label = t.getLabel();
						
						// if it is not an invisible transition, then add it 
						if(!invisibleLabels.contains(label)) efTransitions.add(label);
						
						tobeVisited.add(t.getTarget());
						
						//System.out.println("transition added: " + t.getLabel());
						//System.out.println("tobevisited added: " + t.getTarget().getLabel());
					}				
				}
					
			}
						
			return collectEventuallyFollows(graph, tobeVisited, alreadyVisited, efTransitions);
		}
	}	
	
	public String toString() {
		String behaviouralStr = "";
		
		for(String x : exclusiveness.keySet()) {
			behaviouralStr = behaviouralStr + x + "===";
			Set<String> exclusiveSet = exclusiveness.get(x);
			Iterator<String> it = exclusiveSet.iterator();
			
			while(it.hasNext()) {
				String y = it.next();
				
				behaviouralStr = behaviouralStr + y + "\t";
			}
			
			behaviouralStr += "\n";
		}
		
		behaviouralStr += "=EOF=\n";
		
		for(String x : inverse.keySet()) {
			behaviouralStr = behaviouralStr + x + "===";
			Set<String> inverseSet = inverse.get(x);
			Iterator<String> it = inverseSet.iterator();
			
			while(it.hasNext()) {
				String y = it.next();
				
				behaviouralStr = behaviouralStr + y + "\t";
			}
			
			behaviouralStr += "\n";
		}
		
		behaviouralStr += "=EOF=\n";
		
		return behaviouralStr;
	}
	
	// borrowed from A.Burattion StreamConformance
	public Pair<Petrinet, Marking> unfold(PluginContext context, Petrinet net) throws Exception {
		File pnmlOriginal = File.createTempFile("petrinet", "original");
		String pnmlUnfolded = pnmlOriginal.getAbsolutePath() + "-unfolded";
		
		PnmlExportNetToPNML exporter = new PnmlExportNetToPNML();
		exporter.exportPetriNetToPNMLFile(context, net, pnmlOriginal);
		
		ProcessBuilder pb = new ProcessBuilder("java", "-jar", "./lib/unfolder-v4.jar", pnmlOriginal.getAbsolutePath(), pnmlUnfolded);
		Process proc = pb.start();
		BufferedReader reader = new BufferedReader(new InputStreamReader(proc.getInputStream()));
		while (reader.readLine() != null) { }
		proc.waitFor();
		
		
		Petrinet netUnfolded = context.tryToFindOrConstructFirstNamedObject(
				Petrinet.class,
				PnmlImportNet.class.getAnnotation(Plugin.class).name(),
				null,
				null,
				pnmlUnfolded);
		InitialMarkingConnection imc = context.getConnectionManager().getFirstConnection(InitialMarkingConnection.class, context, netUnfolded);
		
		pnmlOriginal.deleteOnExit();
		return new Pair<Petrinet, Marking>(
				(Petrinet) imc.getObjectWithRole("Net"),
				(Marking) imc.getObjectWithRole("Marking"));
	}

}
