package org.processmining.indatabasepreprocessing.algorithms;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeoutException;

import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryBufferedImpl;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.indatabasepreprocessing.parameters.DBConnectionParameter;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.visualization.PNLogReplayResultVisPanel;

import nl.tue.alignment.TraceByTraceAlignment;
import nl.tue.alignment.TraceReplayTask;

public class InDBPreprocessingAlignment implements InDBPreprocessingAlgorithm {

	private UIPluginContext context;
	private DBConnectionParameter parameter;
	private Statement stat;
	private Petrinet petriNet;
	private Marking initialMarking;
	private Marking finalMarking;
	private XLog log;
	private XEventClasses classes;
	private TransEvClassMapping mapping;
	private HashMap<Integer, int[]> eventsWithErrorsMap;
	private HashMap<String, Integer> caseIdToIndex;
	private HashMap<String, Integer> caseLength;
	
	public InDBPreprocessingAlignment(UIPluginContext context, Connection conn, DBConnectionParameter parameter, Petrinet petriNet) {
		this.context = context;
		this.parameter = parameter;
		this.petriNet = petriNet;
		this.caseIdToIndex = new HashMap<String, Integer>();
		this.eventsWithErrorsMap = new HashMap<Integer, int[]>();
		this.caseLength = new HashMap<String, Integer>();
		try {
			stat = conn.createStatement();
			stat.setFetchSize(100000);
		} catch (SQLException e) {
			e.printStackTrace();
		}		
	}

	public JComponent visualize() {
		PNRepResult repResult = null;
		
		try {
			getMarking(context);
			importLog();
			getMapping(context);
			importErrorIndex();
			repResult = doLogReplay();
		} catch (Exception e) {
			e.printStackTrace();
		}
	
		return new PNLogReplayResultVisPanel(log, repResult, context.getProgress());
	}
	
	private PNRepResult doLogReplay() throws InterruptedException, ExecutionException, TimeoutException {
		TraceByTraceAlignment traceByTraceAlignment = new TraceByTraceAlignment(petriNet, initialMarking, finalMarking, classes, mapping);
	
		List<Future<TraceReplayTask>> list = new ArrayList<>(log.size());
		for (int i=0; i< log.size(); i++) {			
			
			XTrace trace = log.get(i);
			System.out.println("=======TRACE: " + trace.getAttributes().get("concept:name"));
			
			if(eventsWithErrorsMap.get(i) == null) {
				list.add(traceByTraceAlignment.doReplay(log.get(i), i, (15*1000),0));
			} else {
				list.add(traceByTraceAlignment.doReplay(log.get(i), i, (15*1000), 0, eventsWithErrorsMap.get(i)));
			}
		}
				
		PNRepResult repResult = traceByTraceAlignment.merge(list);
		
		Map<String, Object> map = repResult.getInfo();
		for(String key : map.keySet()) {
			System.out.println(key + " " + map.get(key));
		}
		
		return repResult;
	}
	
	private void getMarking(UIPluginContext context) {
		// obtain its initial marking
		try {
			InitialMarkingConnection con = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, petriNet);
			this.initialMarking = (Marking) con.getObjectWithRole(InitialMarkingConnection.MARKING);
			
		} catch (ConnectionCannotBeObtained conProb) {
			JOptionPane.showMessageDialog(new JPanel(),
					"No alignment can be performed as the net does not have any corresponding initial marking");
			context.getFutureResult(0).cancel(true);
			return;
		}
		
		// obtain its final marking
		try {
			FinalMarkingConnection con = context.getConnectionManager()
					.getFirstConnection(FinalMarkingConnection.class, context, petriNet);
			this.finalMarking = (Marking) con.getObjectWithRole(FinalMarkingConnection.MARKING);
			
		} catch (ConnectionCannotBeObtained conProb) {
			JOptionPane.showMessageDialog(new JPanel(),
					"No alignment can be performed as the net does not have any corresponding final marking");
			context.getFutureResult(0).cancel(true);
			return;
		}
	}
	
	private void importErrorIndex() throws SQLException {
		
		ResultSet rs = stat.executeQuery("select * from errorlog");
		
		// take the error index from ERROR LOG table
		HashMap<Integer, ArrayList<Integer>> eventsWithErrorsMapTemp = new HashMap<Integer, ArrayList<Integer>>();
		while(rs.next()) {
			String caseid = rs.getString(1);
			int indexError = rs.getInt(2) + 1;
			
			if(indexError < caseLength.get(caseid)) {
				Integer caseidIndex = caseIdToIndex.get(caseid);
				ArrayList<Integer> error;
				if(eventsWithErrorsMapTemp.containsKey(caseidIndex)) {
					error = eventsWithErrorsMapTemp.get(caseidIndex);
				} else {
					error = new ArrayList<Integer>();
				}
				error.add(indexError);
				eventsWithErrorsMapTemp.put(caseidIndex, error);
			}
			
		}
		
		// put the information into eventsWithErrorMap
		for(Integer caseidIndex : eventsWithErrorsMapTemp.keySet()) {
			ArrayList<Integer> error = eventsWithErrorsMapTemp.get(caseidIndex);
			Collections.sort(error);
			int[] errorArray = new int[error.size()];
			for(int i = 0; i < error.size(); i++) {
					errorArray[i] = error.get(i);
			}
			eventsWithErrorsMap.put(caseidIndex, errorArray);
		}		
	}
	
	private void importLog() throws Exception {
		
		XFactory factory = new XFactoryBufferedImpl();
		ResultSet rs = stat.executeQuery("select * from "+ parameter.getLogTable());
		int caseindex = 0;
		HashMap<String, ArrayList<Event>> cases = new HashMap<String, ArrayList<Event>>();		
		
		while(rs.next()) {
			String caseid = rs.getString(1);
			String activity = rs.getString(2);
			Timestamp time = rs.getTimestamp(3);
			
			ArrayList<Event> events;
			if(cases.containsKey(caseid)) {
				events = cases.get(caseid);
			} else {
				events = new ArrayList<Event>();
			}			
			events.add(new Event(time.getTime(), activity));
			cases.put(caseid, events);
		}
		
		log = factory.createLog();
		
		// add empty trace
		XTrace emptytrace = factory.createTrace();
		emptytrace.getAttributes().put("concept:name", factory.createAttributeLiteral("concept:name", "empty", XConceptExtension.instance()));
		log.add(emptytrace);
		caseIdToIndex.put("empty", caseindex);
		caseindex++;	
		
		for(String caseid : cases.keySet()) {
			XTrace trace = factory.createTrace();
			trace.getAttributes().put("concept:name", factory.createAttributeLiteral("concept:name", caseid, XConceptExtension.instance()));
			
			ArrayList<Event> events = cases.get(caseid);
			Collections.sort(events);
			int i;
			for(i = 0; i < events.size(); i++) {
				XEvent event = factory.createEvent();
				event.getAttributes().put("concept:name", factory.createAttributeLiteral("concept:name", events.get(i).label, XConceptExtension.instance()));
				event.getAttributes().put("time:timestamp", factory.createAttributeTimestamp("time:timestamp", events.get(i).time, XTimeExtension.instance()));
				trace.add(event);
			}
			
			caseIdToIndex.put(caseid, caseindex);
			caseindex++;
			caseLength.put(caseid, i);
			log.add(trace);
		}		
	}
	
	private void getMapping(UIPluginContext context) {
		XEventClassifier eventClassifier = new XEventNameClassifier();
		XEventClass dummyEvClass = new XEventClass("DUMMY", 99999);
		XLogInfo summary = XLogInfoFactory.createLogInfo(log, eventClassifier);
		classes = summary.getEventClasses();
		
		mapping = constructMapping(petriNet, log, dummyEvClass, eventClassifier);
		
		/*XEventClassifier eventClassifier = new XEventNameClassifier();
		XLogInfo summary = XLogInfoFactory.createLogInfo(log, eventClassifier);
		classes = summary.getEventClasses();
		
		EvClassLogPetrinetConnectionFactory mappingFactory =  new EvClassLogPetrinetConnectionFactory();
		EvClassLogPetrinetConnection conn = mappingFactory.connect(context, log, petriNet);
		this.mapping = (TransEvClassMapping) conn.getObjectWithRole(EvClassLogPetrinetConnection.TRANS2EVCLASSMAPPING);*/
	}
	
	private TransEvClassMapping constructMapping(PetrinetGraph net, XLog log, XEventClass dummyEvClass,
			XEventClassifier eventClassifier) {
		TransEvClassMapping mapping = new TransEvClassMapping(eventClassifier, dummyEvClass);

		XLogInfo summary = XLogInfoFactory.createLogInfo(log, eventClassifier);

		for (Transition t : net.getTransitions()) {
			for (XEventClass evClass : summary.getEventClasses().getClasses()) {
				String id = evClass.getId();

				if (t.getLabel().equals(id)) {
					mapping.put(t, evClass);
					break;
				}
			}
		}

		return mapping;
	}
	
	private class Event implements Comparable<Event> {
		long time;
		String label;
		
		Event(long time, String label){
			this.time = time;
			this.label = label;
		}
		
		@Override
		public int compareTo(Event e2) {
			if (time - e2.time < 0){
				return -1;
			}else{
				if (time - e2.time > 0){
					return 1;
				}else{
					return 0;
				}
			}
		}
	}

}
