package org.processmining.indatabasepreprocessing.algorithms;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.h2.util.TriggerInDBPreprocessing;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.indatabasepreprocessing.parameters.DBConnectionParameter;
import org.processmining.indatabasepreprocessing.parameters.IntermediateStructure;
import org.processmining.indatabasepreprocessing.parameters.MiningAlgorithm;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;

public class InDBPreprocessingController {
	
	public DBConnectionParameter parameter;
	private UIPluginContext context;
	private Connection conn;
	private Statement stat;
	private InDBPreprocessingInductiveMinerMSD inductiveMiner;
	private InDBPreprocessingFHM fhm;
	private InDBPreprocessingDHM dhm;
	private InDBPreprocessingDeclare declare;
	private InDBPreprocessingAbsoluteHoW absoluteHoW;
	private InDBPreprocessingAbsoluteCausalHoW absoluteCausalHoW;
	private InDBPreprocessingBooleanHoW booleanHoW;
	private InDBPreprocessingBooleanCausalHoW booleanCausalHoW;
	private InDBPreprocessingAlignment alignment;
	private InDBPreprocessingAlignmentBehavioural alignmentBehavioural;
	private InDBPreprocessingTimeInBetween timeInBetween;
	private MiningAlgorithm currentAlgorithm; 
	private MiningAlgorithm[] chosenAlgorithms;
	private Set<IntermediateStructure> existingIntermediateStructures;
	private Set<IntermediateStructure> neededIntermediateStructures;
	private Set<IntermediateStructure> droppedIntermediateStructures;
	private Petrinet petrinet;
	
	public InDBPreprocessingController(UIPluginContext context, DBConnectionParameter parameter, Petrinet petrinet) {
		this.context = context;
		this.parameter = parameter;
		this.petrinet = petrinet;
		this.chosenAlgorithms = parameter.getPMAlgorithms();
		start();
	}
	
	public InDBPreprocessingController(UIPluginContext context, DBConnectionParameter parameter) {
		this.context = context;
		this.parameter = parameter;
		this.chosenAlgorithms = parameter.getPMAlgorithms();
		start();
	}
	
	private void start() {
		establishConnection();
		checkExistingIntermediateStructure(); 
		checkNeededIntermediateStructure();
		checkDroppedIntermediateStructure();
		dropIntermediateStructure();
		if(parameter.getPreviousData() == true) {
			buildIntermediateStructure();
		} else {
			buildEmptyIntermediateStructure();
		}		
		instantiateMiningAlgorithm();		
		enableTrigger();	
	}
	
	private void establishConnection() {
		try {
			Class.forName("org.h2.Driver");
			conn = DriverManager.getConnection("jdbc:h2:tcp://"+parameter.getServer()+"/~/"+parameter.getDatabaseName()+";LOG=0;CACHE_SIZE=2048000000;LOCK_MODE=0;UNDO_LOG=0;REFERENTIAL_INTEGRITY=FALSE;CACHE_TYPE=TQ", 
					parameter.getUsername(), parameter.getPassword());
			//conn = DriverManager.getConnection("jdbc:h2:mem:", "sa", "");
			stat = conn.createStatement();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (SQLException e) {
			e.printStackTrace();
		}		
		
		if(conn == null) {
			JOptionPane.showConfirmDialog(
					new JPanel(),
					"Database connection cannot be made. Please check the configuration again.",
					"No Database Connection", JOptionPane.PLAIN_MESSAGE, JOptionPane.INFORMATION_MESSAGE);
		} 
	}
	
	private void instantiateMiningAlgorithm() {		
		for(int i = 0 ; i < chosenAlgorithms.length; i++) {
			MiningAlgorithm algorithm = chosenAlgorithms[i];
			switch (algorithm) {
				case INDUCTIVE_MINER :
					inductiveMiner = new InDBPreprocessingInductiveMinerMSD(context, conn, parameter);
					break;					
				case FLEXIBLE_HEURISTIC_MINER :
					fhm = new InDBPreprocessingFHM(context, conn, parameter);
					break;
				case DATA_AWARE_HEURISTIC_MINER :
					dhm = new InDBPreprocessingDHM(context, conn, parameter);
					break;
				case DECLARE_MINERFUL:
					declare = new InDBPreprocessingDeclare(context, conn, parameter);
					break;
				case ABSOLUTE_CAUSAL_HANDOVER_OF_WORK:
					absoluteCausalHoW = new InDBPreprocessingAbsoluteCausalHoW(context, conn, parameter);
					break;
				case ABSOLUTE_HANDOVER_OF_WORK :
					absoluteHoW = new InDBPreprocessingAbsoluteHoW(context, conn, parameter);
					break;
				case BOOLEAN_CAUSAL_HANDOVER_OF_WORK :
					booleanCausalHoW = new InDBPreprocessingBooleanCausalHoW(context, conn, parameter);
					break;
				case BOOLEAN_HANDOVER_OF_WORK :
					booleanHoW = new InDBPreprocessingBooleanHoW(context, conn, parameter);
					break;
				case ALIGNMENT_RULE_CHECKING :
					if(petrinet == null) {
						JOptionPane.showMessageDialog(new JPanel(),
								"No rule checking can be performed as the net does not exist. Use another plug-in.");
						context.getFutureResult(0).cancel(true);
						return;
					}
					alignment = new InDBPreprocessingAlignment(context, conn, parameter, petrinet);
					break;
				case ALIGNMENT_BEHAVIOURAL :
					if(petrinet == null) {
						JOptionPane.showMessageDialog(new JPanel(),
								"No behavioural profiles checking can be performed as the net does not exist. Use another plug-in.");
						context.getFutureResult(0).cancel(true);
						return;
					}
					alignmentBehavioural = new InDBPreprocessingAlignmentBehavioural(context, conn, parameter, petrinet);
					break;
				case PERFORMANCE_ANALYSIS : 
					timeInBetween = new InDBPreprocessingTimeInBetween(context, conn, parameter);
					break;
				case PERFORMANCE_ANALYSIS_NOCHECK : 
					timeInBetween = new InDBPreprocessingTimeInBetween(context, conn, parameter);
					break;
					
			}
		}
		setCurrentAlgorithm(chosenAlgorithms[0]);
	}
	
	private void checkExistingIntermediateStructure() {
		
		existingIntermediateStructures = new HashSet<IntermediateStructure>();
		
		ResultSet rs;
		try {
			rs = stat.executeQuery("show columns from dfr");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.DFR);
			
			rs = stat.executeQuery("show columns from dfr_performance");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.PERFORMANCE_IS);
			
			rs = stat.executeQuery("show columns from length_two_loop");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.LENGTH_TWO_LOOP);
			
			rs = stat.executeQuery("show columns from msd");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.MSD);
			
			rs = stat.executeQuery("show columns from contrariwise_repetition");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.DECLARE_RELATIONS);
			
			rs = stat.executeQuery("show columns from absolute_causal_how");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.ABSOLUTE_CAUSAL_IS);
			
			rs = stat.executeQuery("show columns from boolean_causal_how");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.BOOLEAN_CAUSAL_IS);
			
			rs = stat.executeQuery("show columns from boolean_how");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.BOOLEAN_IS);
			
			rs = stat.executeQuery("show columns from start_event");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.START_EVENT);
			
			rs = stat.executeQuery("show columns from last_event");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
						
			rs = stat.executeQuery("show columns from controller");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.CONTROLLER);
			
			rs = stat.executeQuery("show columns from occurrence");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.OCCURRENCE);
			
			rs = stat.executeQuery("show columns from eventually_follows");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.EVENTUALLY_FOLLOWS);
			
			rs = stat.executeQuery("show columns from eventually_follows_duplicate");
			if(rs.next()) existingIntermediateStructures.add(IntermediateStructure.EVENTUALLY_FOLLOWS_DUPLICATE);
			
			rs = stat.executeQuery("show columns from behavioural_error");
			if(rs.next()) {
				stat.execute("create alias if not exists TRIGGER_GETPROFILE_INDBPREPROCESSING for \"" +
						TriggerInDBPreprocessing.class.getName() + ".getProfile\"");
				ResultSet rs2 = stat.executeQuery("call " + "TRIGGER_GETPROFILE_INDBPREPROCESSING('TRIGGER_INDBPREPROCESSING')");
				String profileInTrigger = (rs2.first()) ? rs2.getString(1) : "";
				
				// if there is errorlog table and rule set in trigger is the same as the current rule, then the alignment IS exists
				if(parameter.getBehaviouralProfile() != null && profileInTrigger.equals(parameter.getBehaviouralProfile())) {
					existingIntermediateStructures.add(IntermediateStructure.BEHAVIOURAL_INDEX_ERROR);
				} else {
					stat.execute("drop table if exists behavioural_error;");
				}
			}
			
			rs = stat.executeQuery("show columns from errorlog");
			if(rs.next()) {
				stat.execute("create alias if not exists TRIGGER_GETRULE_INDBPREPROCESSING for \"" +
						TriggerInDBPreprocessing.class.getName() + ".getRule\"");
				ResultSet rs2 = stat.executeQuery("call " + "TRIGGER_GETRULE_INDBPREPROCESSING('TRIGGER_INDBPREPROCESSING')");
				String ruleSetInTrigger = (rs2.first()) ? rs2.getString(1) : "";
				
				// if there is errorlog table and rule set in trigger is the same as the current rule, then the alignment IS exists
				if(parameter.getRule() != null && ruleSetInTrigger.equals(parameter.getRule())) {
					existingIntermediateStructures.add(IntermediateStructure.ALIGNMENT_RULE_CHECKING_IS);
				} else {
					stat.execute("drop table if exists errorlog;"); 
					stat.execute("drop table if exists activity_frequency;");
					stat.execute("drop table if exists rule_violated;");
					stat.execute("drop table if exists rule_token_violated;");
				}
			}
			
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
	}
	
	private void checkNeededIntermediateStructure() {
		neededIntermediateStructures = new HashSet<IntermediateStructure>();
		
		for(int i = 0; i < chosenAlgorithms.length; i++) {
			switch(chosenAlgorithms[i]) {
				case INDUCTIVE_MINER : 
					neededIntermediateStructures.add(IntermediateStructure.DFR);
					neededIntermediateStructures.add(IntermediateStructure.START_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.MSD);
					neededIntermediateStructures.add(IntermediateStructure.OCCURRENCE);
					break;
				case FLEXIBLE_HEURISTIC_MINER :
					neededIntermediateStructures.add(IntermediateStructure.DFR);
					neededIntermediateStructures.add(IntermediateStructure.LENGTH_TWO_LOOP);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.START_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.EVENTUALLY_FOLLOWS);
					neededIntermediateStructures.add(IntermediateStructure.OCCURRENCE);
					break;
				case DATA_AWARE_HEURISTIC_MINER :
					neededIntermediateStructures.add(IntermediateStructure.DFR);
					neededIntermediateStructures.add(IntermediateStructure.LENGTH_TWO_LOOP);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.START_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.EVENTUALLY_FOLLOWS_DUPLICATE);
					neededIntermediateStructures.add(IntermediateStructure.OCCURRENCE);
					neededIntermediateStructures.add(IntermediateStructure.CONTROLLER);
					break;
				case DECLARE_MINERFUL:
					neededIntermediateStructures.add(IntermediateStructure.DECLARE_RELATIONS);
					neededIntermediateStructures.add(IntermediateStructure.CONTROLLER);
					neededIntermediateStructures.add(IntermediateStructure.OCCURRENCE);
					neededIntermediateStructures.add(IntermediateStructure.DFR);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					break;
				case ABSOLUTE_CAUSAL_HANDOVER_OF_WORK:
					neededIntermediateStructures.add(IntermediateStructure.ABSOLUTE_CAUSAL_IS);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					break;
				case ABSOLUTE_HANDOVER_OF_WORK:
					neededIntermediateStructures.add(IntermediateStructure.ABSOLUTE_CAUSAL_IS);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					break;
				case BOOLEAN_CAUSAL_HANDOVER_OF_WORK:
					neededIntermediateStructures.add(IntermediateStructure.BOOLEAN_CAUSAL_IS);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					break;
				case BOOLEAN_HANDOVER_OF_WORK:
					neededIntermediateStructures.add(IntermediateStructure.BOOLEAN_IS);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					break;
				case ALIGNMENT_RULE_CHECKING:
					neededIntermediateStructures.add(IntermediateStructure.ALIGNMENT_RULE_CHECKING_IS);
					break;
				case ALIGNMENT_BEHAVIOURAL:
					neededIntermediateStructures.add(IntermediateStructure.OCCURRENCE);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.BEHAVIOURAL_INDEX_ERROR);
					break;
				case PERFORMANCE_ANALYSIS:
					neededIntermediateStructures.add(IntermediateStructure.PERFORMANCE_IS);
					neededIntermediateStructures.add(IntermediateStructure.ALIGNMENT_RULE_CHECKING_IS);
					break;
				case PERFORMANCE_ANALYSIS_NOCHECK:
					neededIntermediateStructures.add(IntermediateStructure.DFR);
					neededIntermediateStructures.add(IntermediateStructure.START_EVENT);
					neededIntermediateStructures.add(IntermediateStructure.LAST_EVENT);
					break;
			}
		}
	}

	private void checkDroppedIntermediateStructure() {
		if(parameter.getPreviousData() == true) {
			droppedIntermediateStructures =  new HashSet<IntermediateStructure>(existingIntermediateStructures);
			droppedIntermediateStructures.removeAll(neededIntermediateStructures); // drop IS which is in the intersection between existing and needed
		} else {
			droppedIntermediateStructures =  new HashSet<IntermediateStructure>(existingIntermediateStructures); // all existing IS are dropped
		}
	}
	
	private void dropIntermediateStructure() {				
		Iterator<IntermediateStructure> it = droppedIntermediateStructures.iterator();
		
		try {
			while(it.hasNext()) {
				switch(it.next()) {
					case DFR: stat.execute("drop table if exists dfr;"); break;
					case MSD: stat.execute("drop table if exists msd;"); break;
					case LENGTH_TWO_LOOP: stat.execute("drop table if exists length_two_loop;"); break;
					case EVENTUALLY_FOLLOWS :
						stat.execute("drop table if exists eventually_follows;");
						stat.execute("drop table if exists controller_eventually_follows;");
						break;
					case EVENTUALLY_FOLLOWS_DUPLICATE :
						stat.execute("drop table if exists eventually_follows_duplicate;");
						break;
					case DECLARE_RELATIONS: 
						stat.execute("drop table if exists no_following;"); 
						stat.execute("drop table if exists no_preceding;"); 
						stat.execute("drop table if exists no_cooccurrence;"); 
						stat.execute("drop table if exists repetition;"); 
						stat.execute("drop table if exists contrariwise_repetition;"); 
						break;
					case ABSOLUTE_CAUSAL_IS: stat.execute("drop table if exists absolute_causal_how;"); break;
					case BOOLEAN_CAUSAL_IS: 
						stat.execute("drop table if exists boolean_causal_how;"); 
						stat.execute("drop table if exists controller_boolean_causal_how;"); 
						break;
					case BOOLEAN_IS: 
						stat.execute("drop table if exists boolean_how;");
						stat.execute("drop table if exists controller_boolean_how;"); 
						break;
					case START_EVENT: stat.execute("drop table if exists start_event;"); break;
					case LAST_EVENT: stat.execute("drop table if exists last_event;"); break;
					case CONTROLLER: stat.execute("drop table if exists controller;"); break;
					case OCCURRENCE:
						stat.execute("drop table if exists occurrence;"); 
						stat.execute("drop table if exists occurrence_per_case;");
						break;
					case BEHAVIOURAL_INDEX_ERROR: stat.execute("drop table if exists behavioural_error;"); break;
					case ALIGNMENT_RULE_CHECKING_IS : 
						stat.execute("drop table if exists errorlog;"); 
						stat.execute("drop table if exists activity_frequency;");
						stat.execute("drop table if exists rule_violated;");
						stat.execute("drop table if exists rule_token_violated;");
						break;
					case PERFORMANCE_IS :
						stat.execute("drop table if exists dfr_performance");
						stat.execute("drop table if exists start_event_performance");
						stat.execute("drop table if exists last_event_performance");
				}
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
	}
	
	private void buildEmptyIntermediateStructure() {
		Iterator<IntermediateStructure> it = neededIntermediateStructures.iterator();
		
		try {
			while(it.hasNext()) {
				switch(it.next()) {
					
					case DFR: 
						stat.execute("create table dfr(act1 varchar(255), act2 varchar(255), freq int, sum_of_time bigint, sum_of_time_square bigint, primary key(act1, act2));"); 
						stat.execute("create table log_sequence(caseid varchar(255), seq int, act varchar(255), primary key(caseid, seq));");
						break;
					case MSD: stat.execute("create table msd(act varchar(255), inbetween varchar(255), freq int, distance int, primary key(act, inbetween));"); break;
					case LENGTH_TWO_LOOP: stat.execute("create table length_two_loop(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2))"); break;			
					case EVENTUALLY_FOLLOWS :
						stat.execute("create table eventually_follows(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						stat.execute("create table controller_eventually_follows(caseid varchar(255), act1 varchar(255), act2 varchar(255), primary key(caseid, act1, act2));");
						break;
					case EVENTUALLY_FOLLOWS_DUPLICATE :
						stat.execute("create table eventually_follows_duplicate(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						break;
					case DECLARE_RELATIONS:
						stat.execute("create table no_following(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						stat.execute("create table no_preceding(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						stat.execute("create table no_cooccurrence(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						stat.execute("create table repetition(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						stat.execute("create table contrariwise_repetition(act1 varchar(255), act2 varchar(255), freq int, primary key(act1, act2));");
						break;
					case ABSOLUTE_CAUSAL_IS: 
						stat.execute("create table absolute_causal_how(act1 varchar(255), act2 varchar(255), res1 varchar(255), res2 varchar(255), freq int, primary key(act1, act2, res1, res2));"); 
						break;
					case BOOLEAN_CAUSAL_IS: 
						stat.execute("create table boolean_causal_how(act1 varchar(255), act2 varchar(255), res1 varchar(255), res2 varchar(255), freq int, primary key(act1, act2, res1, res2));");
						stat.execute("create table controller_boolean_causal_how(caseid varchar(255), act1 varchar(255), act2 varchar(255), res1 varchar(255), res2 varchar(255), primary key(caseid, act1, act2, res1, res2));"); 
						break;
					case BOOLEAN_IS: 
						stat.execute("create table boolean_how(res1 varchar(255), res2 varchar(255), freq int, primary key(res1, res2));"); 
						stat.execute("create table controller_boolean_how(caseid varchar(255), res1 varchar(255), res2 varchar(255), primary key(caseid, res1, res2));");
						break;
					case START_EVENT: stat.execute("create table start_event(caseid varchar(255), act varchar(255), primary key(caseid));"); break;
					case LAST_EVENT: stat.execute("create table last_event(caseid varchar(255), act varchar(255), time bigint, res varchar(255), index int, last_two_act varchar(255), primary key(caseid));"); break;
					case CONTROLLER: 		
						stat.execute("create table controller(caseid varchar(255), act1 varchar(255), act2 varchar(255), freq int, primary key(caseid, act1, act2));"); 
						break;
					case OCCURRENCE:
						stat.execute("create table occurrence(act varchar(255), freq int, primary key(act));");
						stat.execute("create table occurrence_per_case(caseid varchar(255), act varchar(255), freq int, lastIndex int,primary key(caseid, act));");
						break;
					case BEHAVIOURAL_INDEX_ERROR:
						stat.execute("create table behavioural_error(caseid varchar(255), index int, primary key(caseid, index));");
						break;
					case PERFORMANCE_IS:
						stat.execute("create table dfr_performance(act1 varchar(255), act2 varchar(255), freq int, sum_of_time bigint, sum_of_time_square bigint, primary key(act1, act2));");
						stat.execute("create table start_event_performance(caseid varchar(255), act varchar(255), primary key(caseid));");
						stat.execute("create table last_event_performance(caseid varchar(255), act varchar(255), time bigint, primary key(caseid));");
						break;
					case ALIGNMENT_RULE_CHECKING_IS: 
						stat.execute("create table errorlog(caseid varchar(255), index int, primary key(caseid, index));");
						stat.execute("create table rule_violated(caseid varchar(255), rulenumber int, primary key(caseid, rulenumber));");
						stat.execute("create table rule_token_violated(caseid varchar(255), rulenumber int, primary key(caseid, rulenumber));");
						
						HashMap<String, Integer> eventClassToInt = parameter.getEventClassToInt();
						String[] eventClasses = new String[eventClassToInt.size()];
						for(String label : eventClassToInt.keySet()) {
							eventClasses[eventClassToInt.get(label)] = label;
						}
						String query = "create table activity_frequency(caseid varchar(255), ";
						for(int i = 0 ; i < eventClasses.length; i++) {
							query = query + eventClasses[i] + " int, ";
						}
						query = query + "lastindex int, primary key(caseid));";
						stat.execute(query);
						
						break;
				}
			}
		} catch (SQLException e) {
			e.printStackTrace();
		} 

	}

	private void buildIntermediateStructure() {
		Set<IntermediateStructure> tobebuiltIntermediateStructures = new HashSet<IntermediateStructure>(neededIntermediateStructures);
		tobebuiltIntermediateStructures.removeAll(existingIntermediateStructures); // only build non-available intermediate structure	
		Iterator<IntermediateStructure> it = tobebuiltIntermediateStructures.iterator();
		
		String toBeBuiltIS = "";
		while(it.hasNext()) {
			String is = it.next().toString();
			toBeBuiltIS = toBeBuiltIS + is + ";";
		}
		
		String rule = (parameter.getRule() != null) ? parameter.getRule() : "";
		String profile = (parameter.getBehaviouralProfile() != null) ? parameter.getBehaviouralProfile() : "";
		
		System.out.println("QUERY: " + toBeBuiltIS + "\t" + rule + "\t" + profile);
		
		try {
			if(toBeBuiltIS.length() > 0) {
				stat.execute("select * from intermediatestructure ('" + toBeBuiltIS + "','true','" + rule + "','" + profile +"', select * from " + parameter.getLogTable() +")");
			} 
				
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}
	
	private void enableTrigger() {
		Iterator<IntermediateStructure> it = neededIntermediateStructures.iterator();
		
		String enabledTrigger = "";
		while(it.hasNext()) {
			enabledTrigger = enabledTrigger + it.next().toString() + "\t";
		}
		
		String rule = (parameter.getRule() != null) ? parameter.getRule() : "";
		String profile = (parameter.getBehaviouralProfile() != null) ? parameter.getBehaviouralProfile() : "";
		
		try {
			stat.execute("create trigger if not exists TRIGGER_INDBPREPROCESSING before insert on " + parameter.getLogTable() + " for each row call \"" +
					TriggerInDBPreprocessing.class.getName() + "\"");
			stat.execute("drop alias if exists TRIGGER_SET_INDBPREPROCESSING");
			stat.execute("create alias if not exists TRIGGER_SET_INDBPREPROCESSING for \"" +
					TriggerInDBPreprocessing.class.getName() + ".setAlgorithm\"");
			stat.execute("call " + "TRIGGER_SET_INDBPREPROCESSING('TRIGGER_INDBPREPROCESSING', '" + enabledTrigger + "','"+ rule + "','" + profile + "')");
			
		} catch (SQLException e) {
			e.printStackTrace();
		}
		
	}
	
	public void setCurrentAlgorithm(MiningAlgorithm algorithm) {
			currentAlgorithm = algorithm;		
	}
	
	public JComponent visualize() {
		switch(currentAlgorithm) {
			case INDUCTIVE_MINER :
				return inductiveMiner.visualize();
			case FLEXIBLE_HEURISTIC_MINER :
				return fhm.visualize();
			case DATA_AWARE_HEURISTIC_MINER :
				return dhm.visualize();
			case DECLARE_MINERFUL :
				return declare.visualize();
			case ABSOLUTE_CAUSAL_HANDOVER_OF_WORK :
				return absoluteCausalHoW.visualize();
			case ABSOLUTE_HANDOVER_OF_WORK :
				return absoluteHoW.visualize();
			case BOOLEAN_CAUSAL_HANDOVER_OF_WORK :
				return booleanCausalHoW.visualize();
			case BOOLEAN_HANDOVER_OF_WORK :
				return booleanHoW.visualize();			
			case ALIGNMENT_RULE_CHECKING :
				return alignment.visualize();
			case ALIGNMENT_BEHAVIOURAL :
				return alignmentBehavioural.visualize();
			case PERFORMANCE_ANALYSIS :
				return timeInBetween.visualize();
			case PERFORMANCE_ANALYSIS_NOCHECK :
				return timeInBetween.visualize();
		}
		return null;
	}
	
}
