package org.processmining.indatabasepreprocessing.algorithms;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import javax.swing.JComponent;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.plugins.VisualizeAcceptingPetriNetPlugin;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.packages.PackageManager.Canceller;
import org.processmining.indatabasepreprocessing.parameters.DBConnectionParameter;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTree;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce.ReductionFailedException;
import org.processmining.plugins.InductiveMiner.efficienttree.UnknownTreeNodeException;
import org.processmining.plugins.inductiveminer2.helperclasses.graphs.IntGraph;
import org.processmining.plugins.inductiveminer2.plugins.InductiveMinerPlugin;
import org.processmining.plugins.inductiveminer2.withoutlog.InductiveMinerWithoutLog;
import org.processmining.plugins.inductiveminer2.withoutlog.dfgmsd.DfgMsd;
import org.processmining.plugins.inductiveminer2.withoutlog.dfgmsd.DfgMsdImpl;
import org.processmining.plugins.inductiveminer2.withoutlog.variants.MiningParametersIMWithoutLog;

@SuppressWarnings("deprecation")
public class InDBPreprocessingInductiveMinerMSD implements InDBPreprocessingAlgorithm {

	private UIPluginContext context;
	private DBConnectionParameter parameter;
	private Statement stat;
	private AcceptingPetriNet petriNet;
	private DfgMsd dfgMsd;
	private HashMap<String, Integer> intActivities;
	
	public InDBPreprocessingInductiveMinerMSD(UIPluginContext context, Connection conn, DBConnectionParameter parameter) {
		this.context = context;
		this.parameter = parameter;
		try {
			stat = conn.createStatement();
			stat.setFetchSize(100000);
		} catch (SQLException e) {
			e.printStackTrace();
		}		
	}

	public JComponent visualize() {
		long start = System.nanoTime();
		
		// get DFG
		long startRetrieve = System.nanoTime();
		try {
			getActivities();
			getDfgMsd();
		} catch (SQLException e) {
			e.printStackTrace();
		}
		System.out.print("IMdm," + String.format("%.3f", (System.nanoTime() - startRetrieve) / 1000000.0) + ";" );
		
		// only for checking DFG and MSD (can be deleted after experiment)
		try {
			print(dfgMsd, parameter.getLogTable());
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		
		// mine
		long startMining = System.nanoTime();
		try {
			mine();
		} catch (UnknownTreeNodeException | ReductionFailedException e) {
			e.printStackTrace();
		} 
		System.out.print(String.format("%.3f", (System.nanoTime() - startMining) / 1000000.0) + ";" );
		
		// visualization
		long startVisualization = System.nanoTime();
		JComponent component = VisualizeAcceptingPetriNetPlugin.visualize(context, petriNet);
		System.out.print(String.format("%.3f", (System.nanoTime() - startVisualization) / 1000000.0) + ";" );
		
		System.out.print(String.format("%.3f", (System.nanoTime() - start) / 1000000.0) + "\n" );
		
		return component;
	}
	
	private void mine() throws UnknownTreeNodeException, ReductionFailedException {
		// mine process tree
		MiningParametersIMWithoutLog parameter = new MiningParametersIMWithoutLog();
		EfficientTree tree =  InductiveMinerWithoutLog.mineEfficientTree(dfgMsd, parameter, new Canceller() {
			public boolean isCancelled() {
				return context.getProgress().isCancelled();
			}
		});
		
		// convert to petri net
		petriNet = InductiveMinerPlugin.postProcessTree2PetriNet(tree, new Canceller() {
			public boolean isCancelled() {
				return context.getProgress().isCancelled();
			}
		});
	}
	
	private void getActivities() throws SQLException {		
		intActivities = new HashMap<String, Integer>();
		
		ResultSet rs = stat.executeQuery("select act from occurrence");
	
		int indexActivity = 0;
		while(rs.next()) {
			intActivities.put(rs.getString(1), indexActivity++);
		}
	}
	
	private void getDfgMsd() throws SQLException {
		Set<String> setActivities = intActivities.keySet();
		String[] arrActivities = setActivities.toArray(new String[setActivities.size()]);
		
		dfgMsd =new DfgMsdImpl(arrActivities);
		
		// PAIR of activities (DFG)
		ResultSet rs = stat.executeQuery("select * from dfr");
		
		while(rs.next()) {
			String act1 = rs.getString(1);
			String act2 = rs.getString(2);
			int freq = rs.getInt(3);
			
			int intAct1 = intActivities.get(act1);
			int intAct2 = intActivities.get(act2);
			
			dfgMsd.addActivity(intAct1);
			dfgMsd.addActivity(intAct2);
			dfgMsd.getDirectlyFollowsGraph().addEdge(intAct1, intAct2, freq);
		}
		
		// PAIR of activities (MSD)
		 rs = stat.executeQuery("select * from msd");
			
		while(rs.next()) {
			String act1 = rs.getString(1);
			String act2 = rs.getString(2);
			int freq = rs.getInt(3);
			
			int intAct1 = intActivities.get(act1);
			int intAct2 = intActivities.get(act2);
			
			dfgMsd.addActivity(intAct1);
			dfgMsd.addActivity(intAct2);
			dfgMsd.getMinimumSelfDistanceGraph().addEdge(intAct1, intAct2, freq);
		}
		
		//START activities
		rs = stat.executeQuery("select act, count(*) from start_event group by act");
		
		while(rs.next()) {
			String act = rs.getString(1);
			int freq = rs.getInt(2);
			
			int intAct = intActivities.get(act);
			
			dfgMsd.addActivity(intAct);
			dfgMsd.getStartActivities().add(intAct, freq);
		}
		
		//LAST activities
		rs = stat.executeQuery("select act, count(*) from last_event group by act");
		
		while(rs.next()) {
			String act = rs.getString(1);
			int freq = rs.getInt(2);
			
			int intAct = intActivities.get(act);
			
			dfgMsd.addActivity(intAct);
			dfgMsd.getEndActivities().add(intAct, freq);
		}
	}
	
	private void print(DfgMsd dfgmsd, String log) throws IOException {
		
		String fileDfg = "C:\\Users\\asyamsiy\\Experiment\\dfg-msd\\dfg-"+log+"2.txt";
		String fileMsd = "C:\\Users\\asyamsiy\\Experiment\\dfg-msd\\msd-"+log+"2.txt";
		
		BufferedWriter writerDfg = new BufferedWriter(new FileWriter(fileDfg));
		BufferedWriter writerMsd = new BufferedWriter(new FileWriter(fileMsd));
		
		IntGraph dfg = dfgmsd.getDirectlyFollowsGraph();
		
		Iterable<Long> edges = dfg.getEdges();
		Iterator<Long> itEdges = edges.iterator();
		while(itEdges.hasNext()) {
			long edge = itEdges.next();
			int source = dfg.getEdgeSource(edge);
			int target = dfg.getEdgeTarget(edge);			
			
			writerDfg.write(dfgmsd.getActivityOfIndex(source) + "--" + dfgmsd.getActivityOfIndex(target) + "\t" + dfg.getEdgeWeight(edge) + "\n");
		}
				
		IntGraph msd = dfgmsd.getMinimumSelfDistanceGraph();
		
		edges = msd.getEdges();
		itEdges = edges.iterator();
		while(itEdges.hasNext()) {
			long edge = itEdges.next();
			int source = msd.getEdgeSource(edge);
			int target = msd.getEdgeTarget(edge);			
			
			writerMsd.write(dfgmsd.getActivityOfIndex(source) + "--" + dfgmsd.getActivityOfIndex(target) + "\t" + msd.getEdgeWeight(edge) + "\n");
		}
		
		writerDfg.close();
		writerMsd.close();
	}
}
