package org.processmining.indatabasepreprocessing.algorithms;

import java.awt.Color;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.HashMap;

import javax.swing.JComponent;
import javax.swing.SwingConstants;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.util.ui.scalableview.ScalableViewPanel;
import org.processmining.framework.util.ui.scalableview.interaction.ExportInteractionPanel;
import org.processmining.framework.util.ui.scalableview.interaction.PIPInteractionPanel;
import org.processmining.framework.util.ui.scalableview.interaction.ZoomInteractionPanel;
import org.processmining.indatabasepreprocessing.parameters.DBConnectionParameter;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.ProMJGraph;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.plugins.InductiveMiner.dfgOnly.Dfg;
import org.processmining.plugins.InductiveMiner.dfgOnly.DfgImpl;
import org.processmining.plugins.InductiveMiner.dfgOnly.DfgMiningParameters;
import org.processmining.plugins.InductiveMiner.dfgOnly.DfgMiningParametersIMd;
import org.processmining.plugins.InductiveMiner.dfgOnly.plugins.IMdProcessTree;
import org.processmining.plugins.manifestanalysis.visualization.performance.PerformanceConstants;
import org.processmining.plugins.manifestanalysis.visualization.performance.TimeFormatter;
import org.processmining.plugins.manifestanalysis.visualization.performance.TransPerfDecorator;
import org.processmining.plugins.petrinet.PetriNetVisualization;
import org.processmining.plugins.pnalignanalysis.visualization.projection.util.GraphBuilder;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.conversion.ProcessTree2Petrinet;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.InvalidProcessTreeException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.NotYetImplementedException;
import org.processmining.processtree.conversion.ProcessTree2Petrinet.PetrinetWithMarkings;

public class InDBPreprocessingTimeInBetween implements InDBPreprocessingAlgorithm {
	
	private UIPluginContext context;
	private DBConnectionParameter parameter;
	private Statement stat;
	private Petrinet petriNet;
	private Dfg dfg;
	private HashMap<String, Integer> freqMap;
	private HashMap<String, Long> sumOfTimeMap;
	private HashMap<String, Long> sumOfTimeSquareMap;
	private HashMap<String, Double> avgMap;
	private ProMJGraph graph;
	
	public InDBPreprocessingTimeInBetween(UIPluginContext context, Connection conn, DBConnectionParameter parameter) {
		this.context = context;
		this.parameter = parameter;
		try {
			stat = conn.createStatement();
			stat.setFetchSize(100000);
		} catch (SQLException e) {
			e.printStackTrace();
		}		
	}

	public JComponent visualize() {
		long start = System.nanoTime();
		
		// get DFG, Freq, and Time
		long startRetrieve = System.nanoTime();
		try {
			getDfgAndTime();
		} catch (SQLException e) {
			e.printStackTrace();
		}
		System.out.print("Time," + String.format("%.3f", (System.nanoTime() - startRetrieve) / 1000.0) + ";" );
		
		// mine
		long startMining = System.nanoTime();
		try {
			mine();
		} catch (NotYetImplementedException | InvalidProcessTreeException e) {
			e.printStackTrace();
		}	
		System.out.print(String.format("%.3f", (System.nanoTime() - startMining) / 1000.0) + ";" );
		
		// visualization (plus enhancement)
		long startVisualization = System.nanoTime();		
		PetriNetVisualization pnVis = new PetriNetVisualization();
		pnVis.visualize(context, petriNet);		
		try {
			displayAverageTime();
		} catch (ConnectionCannotBeObtained e) {
			e.printStackTrace();
		}		
		ProMJGraphPanel panel = new ProMJGraphPanel(graph);				
		panel.addViewInteractionPanel(new PIPInteractionPanel(panel), SwingConstants.NORTH);
		panel.addViewInteractionPanel(new ZoomInteractionPanel(panel, ScalableViewPanel.MAX_ZOOM), SwingConstants.WEST);
		panel.addViewInteractionPanel(new ExportInteractionPanel(panel), SwingConstants.SOUTH);		
		System.out.print(String.format("%.3f", (System.nanoTime() - startVisualization) / 1000.0) + ";" );
		
		System.out.print(String.format("%.3f", (System.nanoTime() - start) / 1000.0) + "\n" );
		
		return panel;
	}
	
	private void displayAverageTime() throws ConnectionCannotBeObtained {
		GraphLayoutConnection layoutConn = context.getConnectionManager().getFirstConnection(GraphLayoutConnection.class, context, petriNet);		
		graph = GraphBuilder.buildJGraph(petriNet, layoutConn);
		avgMap = new HashMap<String, Double>();
		
		for(Transition t : petriNet.getTransitions()) {			
			String label = t.getLabel();
			Long sum = (sumOfTimeMap.containsKey(label)) ? sumOfTimeMap.get(label) : null;
			Integer freq = (freqMap.containsKey(label)) ? freqMap.get(label) : null;
			if(sum != null) {
				Double avg = (double) (sum/freq);
				avgMap.put(label, avg);
			} else {
				avgMap.put(label, Double.NaN);
			}
		}
		
		changeTransitionColor();
		
		/*graph.addGraphSelectionListener(new GraphSelectionListener() {
			public void valueChanged(GraphSelectionEvent e) {
				// selection of a transition would change the stats
				if (e.getCell() instanceof ProMGraphCell) {
					DirectedGraphNode cell = ((ProMGraphCell) e.getCell()).getNode();
					if (cell instanceof Transition) {
						System.out.println("average: " + cell.getLabel() + " " +  sumOfTimeMap.get(cell.getLabel())/freqMap.get(cell.getLabel()));
					} 
				}
			}
		});*/
	}
	
	private void changeTransitionColor() {
		
		ViewSpecificAttributeMap vsMap = graph.getViewSpecificAttributes();
		NumberFormat nfDouble = NumberFormat.getInstance();
		
		// identify minimum and maximum value
		double[] temp = new double[petriNet.getTransitions().size()];
		Arrays.fill(temp, Double.MAX_VALUE);
		int pointer = 0;

		// find limits
		for (Transition t : petriNet.getTransitions()) {
			if (!t.isInvisible()) {
				String label = t.getLabel();
				double avg = avgMap.get(label);
				if (!Double.isNaN(avg)) {
					temp[pointer++] = avg;
				}
			}
		}

		Arrays.sort(temp);

		for (Transition t : petriNet.getTransitions()) {
			TransPerfDecorator dec = new TransPerfDecorator();
			
			if (!t.isInvisible()) {
				double avg = avgMap.get(t.getLabel());
				Color color = getAppropriateFillColoring(temp[0], !Double.isNaN(avg) ? temp[pointer - 1]
						: temp[0], avg);
				vsMap.putViewSpecific(t, AttributeMap.FILLCOLOR, color);

				dec.setLightColorLabel(adjustLightDarkTextColor(color));
				dec.setStats(TimeFormatter.formatTime(avg, nfDouble));
				
			}
			
			vsMap.putViewSpecific(t, AttributeMap.SHAPEDECORATOR, dec);
		}
	}
	
	private boolean adjustLightDarkTextColor(Color color) {
		return (color.equals(PerformanceConstants.BAD) || color.equals(PerformanceConstants.BADTOMEDIUM));
	}
	
	
	
	private Color getAppropriateFillColoring(double min, double max, double value) {
		if (Double.isNaN(value)) {
			return new Color(255, 255, 255, 0);
		} else
		// get color in between
		if ((Double.compare(max, 0.0) > 0) && (!Double.isNaN(max))) {
			int res = (int) ((value - min) * 5 / (max - min));
			if (res == 0) {
				return PerformanceConstants.GOOD;
			} else if (res == 1) {
				return PerformanceConstants.MEDTOGOOD;
			} else if (res == 2) {
				return PerformanceConstants.MEDIUM;
			} else if (res == 3) {
				return PerformanceConstants.BADTOMEDIUM;
			} else {
				return PerformanceConstants.BAD;
			}
		} else {
			return PerformanceConstants.GOOD;
		}
	}
		
	@SuppressWarnings("deprecation")
	private void mine() throws NotYetImplementedException, InvalidProcessTreeException {
		// mine process tree
		DfgMiningParameters param = new DfgMiningParametersIMd();
		ProcessTree processtree = IMdProcessTree.mineProcessTree(dfg, param);
		
		// convert to petri net
		ProcessTree2Petrinet pt2pn = new ProcessTree2Petrinet();
		PetrinetWithMarkings result = pt2pn.convert(processtree, false);
		
		petriNet = result.petrinet;
	}
	
	
	private void getDfgAndTime() throws SQLException {
		freqMap = new HashMap<String, Integer>();
		sumOfTimeMap = new HashMap<String, Long>();
		sumOfTimeSquareMap = new HashMap<String, Long>();
		
		dfg = new DfgImpl();
		HashMap<String,Integer> eventClassIndex = new HashMap<String, Integer>();
		int index = 0;
		
		// PAIR of activities
		ResultSet rs = stat.executeQuery("select * from dfr"); // change from dfr_performance
		
		while(rs.next()) {
			String elem1 = rs.getString(1);
			String elem2 = rs.getString(2);
			int freq = rs.getInt(3);
			long sumOfTime = rs.getLong(4);
			long sumOfTimeSquare = rs.getLong(5);
			
			int freqElem = (freqMap.containsKey(elem2)) ? freqMap.get(elem2) : 0;
			freqMap.put(elem2, freqElem + freq);
			
			long prevSumOfTime = (sumOfTimeMap.containsKey(elem2)) ? sumOfTimeMap.get(elem2) : 0;
			sumOfTimeMap.put(elem2, prevSumOfTime + sumOfTime);

			long prevSumOfTimeSquare = (sumOfTimeSquareMap.containsKey(elem2)) ? sumOfTimeSquareMap.get(elem2) : 0;
			sumOfTimeSquareMap.put(elem2, prevSumOfTimeSquare + sumOfTimeSquare);
										
			if(!eventClassIndex.containsKey(elem1)) {
				eventClassIndex.put(elem1, index);
				index++;
			}				
			if(!eventClassIndex.containsKey(elem2)) {
				eventClassIndex.put(elem2, index);
				index++;
			}
			
			XEventClass ec1 = new XEventClass(elem1, eventClassIndex.get(elem1));
			XEventClass ec2 = new XEventClass(elem2, eventClassIndex.get(elem2));					
			
			dfg.addActivity(ec1);
			dfg.addActivity(ec2);
			dfg.addDirectlyFollowsEdge(ec1, ec2, freq);			
		}
		
		//START activities
		rs = stat.executeQuery("select act, count(*) from start_event group by act"); // change from start_event_performance
		
		while(rs.next()) {
			String eventLabel = rs.getString(1);
			int freq = rs.getInt(2);
							
			if(!eventClassIndex.containsKey(eventLabel)) {
				eventClassIndex.put(eventLabel, index);
				index++;
			}				
			
			XEventClass startEventLabel = new XEventClass(eventLabel, eventClassIndex.get(eventLabel));	
			
			dfg.addActivity(startEventLabel);
			dfg.addStartActivity(startEventLabel, freq);
		}
		
		//LAST activities
		rs = stat.executeQuery("select act, count(*) from last_event group by act"); // change from last_event_performance
		
		while(rs.next()) {
			String eventLabel = rs.getString(1);
			int freq = rs.getInt(2);
							
			if(!eventClassIndex.containsKey(eventLabel)) {
				eventClassIndex.put(eventLabel, index);
				index++;
			}				
			
			XEventClass lastEventLabel = new XEventClass(eventLabel, eventClassIndex.get(eventLabel));	
			
			dfg.addActivity(lastEventLabel);
			dfg.addEndActivity(lastEventLabel, freq);
		}
	}

}
