package org.processmining.indatabasepreprocessing.algorithms;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.processmining.indatabasepreprocessing.parameters.PlaceObject;
import org.processmining.indatabasepreprocessing.parameters.Rule;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

public class RuleProduction {
	private Rule rule;
	private Petrinet petrinet;
	private Set<String> duplicateLabels;
	private ArrayList<Transition> invisibleTransitions;
	private Set<Place> placesOfInvisibleTransitions;
	private Set<String> invisibleLabels;
	private HashMap<String, ArrayList<Transition>> transitionNodes;
	private HashMap<Place, Integer> initialPlaces;
	private HashMap<Transition, ArrayList<Place>> placeBackward = new HashMap<Transition, ArrayList<Place>>();
	private HashMap<Transition, ArrayList<Place>> placeForward = new HashMap<Transition, ArrayList<Place>>();
	private HashMap<Place,PlaceObject> placeObjects = new HashMap<Place,PlaceObject>();
	private HashMap<String, ArrayList<ArrayList<Place>>> ruleTypeAndCombinedPlaces = new HashMap<String, ArrayList<ArrayList<Place>>>();
	private Integer globalIndex;
	private Integer indexRule;
	private Integer indexRuleToken;
	
	/**
	 * the constructor of RuleProduction.
	 * first it creates map from transition-label to transition nodes, then detect the duplicate labels.
	 * next it handles initial marking, create rule, and finally create rule for combination.
	 * 
	 * @param petrinet
	 * @param marking
	 */
	public RuleProduction(Petrinet petrinet, Marking marking) {		
		this.rule = new Rule();
		this.petrinet = petrinet;
		this.transitionNodes = new HashMap<String,ArrayList<Transition>>();
		this.duplicateLabels = new HashSet<String>();
		this.initialPlaces = new HashMap<Place, Integer>();
		this.invisibleTransitions = new ArrayList<Transition>();
		this.placesOfInvisibleTransitions = new HashSet<Place>();
		this.invisibleLabels = new HashSet<String>();
		this.invisibleLabels.add("$invisible$"); this.invisibleLabels.add("tau start"); this.invisibleLabels.add("tau join"); this.invisibleLabels.add("tau split"); this.invisibleLabels.add("tau from tree");
		this.invisibleLabels.add("t5"); this.invisibleLabels.add("t6"); this.invisibleLabels.add("t7"); this.invisibleLabels.add("t8"); this.invisibleLabels.add("t11"); this.invisibleLabels.add("t12"); this.invisibleLabels.add("t14"); this.invisibleLabels.add("t15");
		this.globalIndex = 0;	
		this.indexRule = 0;
		this.indexRuleToken = 0;		
		
		// create hashmap transition-label to transition-nodes	and set of duplicate labels
		Collection<Transition> transitions = petrinet.getTransitions();
		Iterator<Transition> itTransitions = transitions.iterator();
		
		// iterate transitions
		while(itTransitions.hasNext()) {
			Transition t = itTransitions.next();
			String label = t.getLabel();
			
			// take duplicate labels
			ArrayList<Transition> list;
			if(transitionNodes.containsKey(label) && !invisibleLabels.contains(label)) {
				list = transitionNodes.get(label);
				duplicateLabels.add(label);
			} else {
				list = new ArrayList<Transition>();				
			}
			list.add(t);
			transitionNodes.put(label, list);
			
			// take invisible transitions
			if(invisibleLabels.contains(label)) {
				invisibleTransitions.add(t);
			}
		}		
		
		// take the initial places		
		Iterator<Place> itMarking = marking.iterator();
		while(itMarking.hasNext()) {
			Place p = itMarking.next();
			 
			if(initialPlaces.containsKey(p)) {
				int token = initialPlaces.get(p);
				initialPlaces.put(p, token+1);
			} else {
				initialPlaces.put(p, 1);
			}			
		}
		
		// create rule
		createRule();
		
		// handle invisible transition
		Iterator<Place> itPlaceOfInvisibleTransitions = placesOfInvisibleTransitions.iterator();
		while(itPlaceOfInvisibleTransitions.hasNext()) {
			Place place = itPlaceOfInvisibleTransitions.next();
			//System.out.println("place: " + place.getLabel());
				
			Object[] arr = invisibleOnLeft(place);
				
			ArrayList<ArrayList<String>> leftLabelsCollection = (ArrayList<ArrayList<String>>) arr[0];
			ArrayList<ArrayList<String>> rightLabelsCollection = (ArrayList<ArrayList<String>>) arr[1];
			
			if(!rightLabelsCollection.isEmpty()) {
				for(int j = 0; j < leftLabelsCollection.size(); j++) {				
					ArrayList<String> leftLabel = leftLabelsCollection.get(j);
					ArrayList<Integer> leftSideRule = new ArrayList<Integer>();
					String leftLabelStr = "";
					Integer intLabel;
					for(int k = 0; k < leftLabel.size(); k++) {
						String label = leftLabel.get(k);
						leftLabelStr = leftLabelStr + label + " ";
						if(rule.eventClassToInt.containsKey(label)) {
							intLabel = rule.eventClassToInt.get(label);
						} else {
							intLabel = globalIndex;
							rule.eventClassToInt.put(label, intLabel);
							globalIndex++;
						}
						leftSideRule.add(intLabel);
					}
					
					for(int l = 0; l < rightLabelsCollection.size(); l++) {
						ArrayList<String> rightLabel = rightLabelsCollection.get(l);
						ArrayList<Integer> rightSideRule = new ArrayList<Integer>();
						String rightLabelStr = "";
						Integer intLabelR;
						for(int m = 0; m < rightLabel.size(); m++) {
							String label = rightLabel.get(m);
							rightLabelStr = rightLabelStr + label + " ";
							if(rule.eventClassToInt.containsKey(label)) {
								intLabelR = rule.eventClassToInt.get(label);
							} else {
								intLabelR = globalIndex;
								rule.eventClassToInt.put(label, intLabelR);
								globalIndex++;
							}
							rightSideRule.add(intLabelR);
						}
						
						AbstractMap.SimpleEntry<ArrayList<Integer>,ArrayList<Integer>> pair = new AbstractMap.SimpleEntry<ArrayList<Integer>,ArrayList<Integer>>(leftSideRule, rightSideRule);
						rule.rules.put(indexRule, pair);
						
						// create a map between activity to rule associated with it
						for(int n = 0; n < rightSideRule.size(); n++) {
							Integer actInt = rightSideRule.get(n);
							ArrayList<Integer> indexesRules;
							if(rule.rulesOfaLabel.containsKey(actInt)) {
								indexesRules = rule.rulesOfaLabel.get(actInt);							
							} else {
								indexesRules = new ArrayList<Integer>();
							}
							indexesRules.add(indexRule);
							rule.rulesOfaLabel.put(actInt, indexesRules);
						}
						
						indexRule++;						
						//System.out.println("rule invisible : " + leftLabelStr + " >= " + rightLabelStr);
					}
				}
			}
			
		}
		
		// create rule for duplicate labels
		if(!duplicateLabels.isEmpty()) {
			createRuleForDuplicateLabels();
		}

	}
	
	private Object[] invisibleOnLeft(Place place) {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> leftTransitions = petrinet.getInEdges(place); 
		Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itLeftTransitions = leftTransitions.iterator();
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> rightTransitions = petrinet.getOutEdges(place); 
		Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itRightTransitions = rightTransitions.iterator();
		ArrayList<ArrayList<String>> collectionLeftLabels = new ArrayList<ArrayList<String>>();
		ArrayList<ArrayList<String>> collectionRightLabels = new ArrayList<ArrayList<String>>();
		
		// left side of a place
		while(itLeftTransitions.hasNext()) {
			Transition t = (Transition) itLeftTransitions.next().getSource();
			String label = t.getLabel();
			if(!invisibleLabels.contains(label)) {
				if(collectionLeftLabels.isEmpty()) {
					ArrayList<String> leftLabels = new ArrayList<String>();
					leftLabels.add(label);
					collectionLeftLabels.add(leftLabels);
				} else {
					for(int i = 0; i < collectionLeftLabels.size(); i++) {
						collectionLeftLabels.get(i).add(label);
					}
				}
			} else {
				
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> leftPlaces = petrinet.getInEdges(t);
				Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itLeftPlaces = leftPlaces.iterator();		
				
				
				ArrayList<ArrayList<String>> collectionLeftLabelsCopy = collectionLeftLabels;
				collectionLeftLabels = new ArrayList<ArrayList<String>>();
				
				while(itLeftPlaces.hasNext()) {
					Place p = (Place) itLeftPlaces.next().getSource();					
					Object[] doubleCollections = invisibleOnLeft(p);
					ArrayList<ArrayList<String>> leftLabelsCollection = (ArrayList<ArrayList<String>>) doubleCollections[0];
					ArrayList<ArrayList<String>> rightLabelsCollection = (ArrayList<ArrayList<String>>) doubleCollections[1];
					//collectionLeftLabels.addAll(leftLabelsCollection);
					collectionRightLabels.addAll(rightLabelsCollection);
					
					ArrayList<ArrayList<String>> tempListForCombined = new ArrayList<ArrayList<String>>();
					
					if(collectionLeftLabelsCopy.isEmpty()) {
						collectionLeftLabels.addAll(leftLabelsCollection);
					} else {
						for(int i = 0; i < collectionLeftLabelsCopy.size(); i++) {
							for(int j = 0; j < leftLabelsCollection.size(); j++) {
								ArrayList<String> list = new ArrayList<String>();
								list.addAll(collectionLeftLabelsCopy.get(i));
								list.addAll(leftLabelsCollection.get(j));
								tempListForCombined.add(list);
							}
						}
						collectionLeftLabels.addAll(tempListForCombined);
					}
					
					//System.out.println("=== place: " + p.getLabel() + " " + collectionLeftLabels.size() + " ");
				}
			}
		}
		
		// right side of a place
		while(itRightTransitions.hasNext()) {
			Transition t = (Transition) itRightTransitions.next().getTarget();
			String label = t.getLabel();
			if(!invisibleLabels.contains(label)) {
				if(collectionRightLabels.isEmpty()) {
					ArrayList<String> rightLabels = new ArrayList<String>();
					rightLabels.add(label);
					collectionRightLabels.add(rightLabels);
				} else {
					for(int i = 0; i < collectionRightLabels.size(); i++) {
						collectionRightLabels.get(i).add(label);
					}
				}
			}
		}
		
		return new Object[] {collectionLeftLabels, collectionRightLabels};
	}
	
	/**
	 * method to create rule without duplication.
	 * simply loop through the place.
	 * 
	 */
	private void createRule() {
		Collection<Place> places = petrinet.getPlaces();
		Iterator<Place> itPlaces = places.iterator();
		
		placeIteration: while(itPlaces.hasNext()) {
			Place p = itPlaces.next();
			ArrayList<Integer> leftSideRule = new ArrayList<Integer>();
			ArrayList<Integer> rightSideRule = new ArrayList<Integer>();
			boolean existsInvisible = false;
			
			// in
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inTransitions = petrinet.getInEdges(p); 
			Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itInTransitions = inTransitions.iterator();
			
			String in = "";
			while(itInTransitions.hasNext()) {
				String label = itInTransitions.next().getSource().getLabel();
				
				if(invisibleLabels.contains(label)) {
					existsInvisible = true;
					placesOfInvisibleTransitions.add(p);
					continue;
				}
				if(duplicateLabels.contains(label)) {
					continue placeIteration;
				} else {
					in = in + " + " + label;
					Integer intLabel;
					if(rule.eventClassToInt.containsKey(label)) {
						intLabel = rule.eventClassToInt.get(label);
					} else {
						intLabel = globalIndex;
						rule.eventClassToInt.put(label, intLabel);
						globalIndex++;
					}
					leftSideRule.add(intLabel);
				}				
			}
			
			// out
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outTransitions = petrinet.getOutEdges(p);
			Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itOutTransitions = outTransitions.iterator();
			
			String out =  "";
			while(itOutTransitions.hasNext()) {
				String label = itOutTransitions.next().getTarget().getLabel();
				
				if(invisibleLabels.contains(label)) {
					continue;
				}
				if(duplicateLabels.contains(label)) {
					continue placeIteration;
				} else {
					out = out + " + " + label;
					Integer intLabel;
					if(rule.eventClassToInt.containsKey(label)) {
						intLabel = rule.eventClassToInt.get(label);
					} else {
						intLabel = globalIndex;
						rule.eventClassToInt.put(label, intLabel);
						globalIndex++;
					}
					rightSideRule.add(intLabel);
				}
			}
			
			// involve marking in initial places
			Integer token = 0;
			if(initialPlaces.containsKey(p)) {
				token = initialPlaces.get(p);
				in = in + " + " + token;
			}
			
			// does not involve final places & does not exist invisible transitions
			if(!rightSideRule.isEmpty() && !existsInvisible) {
				if(token == 0) {
					AbstractMap.SimpleEntry<ArrayList<Integer>,ArrayList<Integer>> pair = new AbstractMap.SimpleEntry<ArrayList<Integer>,ArrayList<Integer>>(leftSideRule, rightSideRule);
					rule.rules.put(indexRule, pair);
					
					// create a map between activity to rule associated with it
					for(int i = 0; i < rightSideRule.size(); i++) {
						Integer actInt = rightSideRule.get(i);
						ArrayList<Integer> indexesRules;
						if(rule.rulesOfaLabel.containsKey(actInt)) {
							indexesRules = rule.rulesOfaLabel.get(actInt);							
						} else {
							indexesRules = new ArrayList<Integer>();
						}
						indexesRules.add(indexRule);
						rule.rulesOfaLabel.put(actInt, indexesRules);
					}
					
					indexRule++;
				} else {
					AbstractMap.SimpleEntry<Integer,ArrayList<Integer>> pairToken = new AbstractMap.SimpleEntry<Integer,ArrayList<Integer>>(token, leftSideRule);
					AbstractMap.SimpleEntry<AbstractMap.SimpleEntry<Integer,ArrayList<Integer>>, ArrayList<Integer>> pair = new AbstractMap.SimpleEntry<AbstractMap.SimpleEntry<Integer,ArrayList<Integer>>, ArrayList<Integer>>(pairToken, rightSideRule);
					rule.rulesWithToken.put(indexRuleToken, pair);
					
					// create a map between activity to rule associated with it
					for(int i = 0; i < rightSideRule.size(); i++) {
						Integer actInt = rightSideRule.get(i);
						ArrayList<Integer> indexesRules;
						if(rule.rulesTokenOfaLabel.containsKey(actInt)) {
							indexesRules = rule.rulesTokenOfaLabel.get(actInt);							
						} else {
							indexesRules = new ArrayList<Integer>();
						}
						indexesRules.add(indexRuleToken);
						rule.rulesTokenOfaLabel.put(actInt, indexesRules);
					}
					
					indexRuleToken++;
				}
				in = in.substring(3);
				out = out.substring(3);
				//System.out.println("rule : " +  in + " >= " + out);
			}
		}		
	}
	
	
	
	/**
	 * method to create rule with duplication labels.
	 * first it creates tree, make place combination, create recursive rule, and finally print the rule.
	 * 
	 */	
	private void createRuleForDuplicateLabels() {
		
		// create tree for rule backward
		createTreeRuleBackward();
		
		// create tree for rule forward
		createTreeRuleForward();
		
		// create rule by combining places (backward)
		Iterator<String> itDuplicateLabels = duplicateLabels.iterator();		
		while(itDuplicateLabels.hasNext()) {
			String label = itDuplicateLabels.next();			
			int length = transitionNodes.get(label).size();
			combinePlaces(placeBackward, label, label + "_backward", new Place[length], 0, length); // resulting, for example: A_backward -> [p1, p3], [p2, p3]
		}
		
		// create rule by combining places (forward)
		itDuplicateLabels = duplicateLabels.iterator();		
		while(itDuplicateLabels.hasNext()) {
			String label = itDuplicateLabels.next();			
			int length = transitionNodes.get(label).size();
			combinePlaces(placeForward, label, label + "_forward", new Place[length], 0, length); // resulting, for example: A_forward -> [p4], [p5, p6]
		}
					
		// create recursive rule (looping on ruleTypeAndCombinedPlaces, example of its member: A_backward -> [p1, p3], [p2, p3])
		for(String ruleType : ruleTypeAndCombinedPlaces.keySet()) { // for each ruleType, e.g. A_backward
			ArrayList<ArrayList<Place>> placeCombinationsList = ruleTypeAndCombinedPlaces.get(ruleType);
			
			for(int i = 0; i < placeCombinationsList.size(); i++) {	// for each placeCombination of the ruleType, e.g. [p1,p3]
				ArrayList<Place> placeCombination = placeCombinationsList.get(i);
				ArrayList<AbstractMap.SimpleEntry<Place,String>> placeCombinationAndRuleType = new ArrayList<AbstractMap.SimpleEntry<Place,String>>();
				
				for(int j = 0; j < placeCombination.size(); j++) { // for each place, e.g. p1, add the ruleType. this results (p1, A_backward)
																						// this is needed to differentiate between p1, A_backward and p1, B_forward for example
					AbstractMap.SimpleEntry<Place,String> p = new AbstractMap.SimpleEntry<Place,String>(placeCombination.get(j), ruleType);
					placeCombinationAndRuleType.add(p);
				}
				
				createRecursiveRule(placeCombinationAndRuleType, new HashSet<Place>(), 0, placeCombinationAndRuleType.size());
			}
		}
		
		// print recursive rule
		printRecursiveRule();
	}
	
	/**
	 * create a rule by recursively go through place combination.
	 * merge a rule with another rule when necessary (related through a link).
	 * 
	 * @param places, it is an array list of pair(place,ruleType)
	 * @param processedPlace, it is a set of places which have been processed
	 * @param start, index for start
	 * @param end, index for end
	 */
	private void createRecursiveRule(ArrayList<AbstractMap.SimpleEntry<Place,String>> places, Set<Place> processedPlace, int start, int end) {
		for(int i = start; i < end && end == places.size(); i++) {
			Place place = places.get(i).getKey();
			String ruleType = places.get(i).getValue();
			PlaceObject placeobject = placeObjects.get(place);
    		ArrayList<String> links = placeobject.links;
    		processedPlace.add(place);
    		
    		// if it is a final place, does not process
    		if(placeobject.transitionOutEdge.isEmpty()) {
				return;
			}
    			
    		// for each link
			for(int j = 0; j < links.size(); j++) {
				String ruleTypeLink = links.get(j);
				
				if(!ruleType.equals(ruleTypeLink)) { // does not process ruleType which has been processed
					
					ArrayList<ArrayList<Place>> ruleList = ruleTypeAndCombinedPlaces.get(ruleTypeLink); // take the set of rules (set of place combination) of the ruleType
					
					for(int k = 0; k < ruleList.size(); k++) { // for each place combination
						ArrayList<Place> addedRule = ruleList.get(k);
						
						if(addedRule.contains(place)) {	// only process rule which relates to the place under review	
							// remove already processed places 
							ArrayList<AbstractMap.SimpleEntry<Place,String>> tempPlaces = new ArrayList<AbstractMap.SimpleEntry<Place,String>>();
					    	tempPlaces = places;
					    	places = new ArrayList<AbstractMap.SimpleEntry<Place,String>>();					    	
				
							for (int l = 0; l < end; l++ ) { // only keep from index 0 till index end
								places.add(tempPlaces.get(l)); 
							}
							for(int l = 0; l < addedRule.size(); l++) { // add new rule (place combination)
								AbstractMap.SimpleEntry<Place,String> p = new AbstractMap.SimpleEntry<Place,String>(addedRule.get(l),ruleTypeLink);
								if(!processedPlace.contains(p.getKey())) {
									places.add(p);
									processedPlace.add(p.getKey());
								}
							}
							
							createRecursiveRule(places, processedPlace, i+1, places.size());  // recursive call
						}						
					}
				}
				
			}
    		
    	}
    	
		if(end == places.size()) {
			//print		
			Set<Place> setofPlaces = new HashSet<Place>();
			for(int i = 0; i < places.size(); i++) {
				Place p = places.get(i).getKey();
				setofPlaces.add(p);
			}
			rule.placesOfARule.add(setofPlaces);
						
			return;
    	}
	}
	
	/**
	 * print the recursive rule
	 * 
	 */
	private void printRecursiveRule() {
		Set<Set<Place>> setOfSetPlaces = rule.placesOfARule;
		Iterator<Set<Place>> itSetOfSetPlaces = setOfSetPlaces.iterator();
		
		while(itSetOfSetPlaces.hasNext()) {
			Set<Place> setPlaces = itSetOfSetPlaces.next();
			Iterator<Place> itSetPlaces = setPlaces.iterator();
			
			ArrayList<Integer> leftSideRule = new ArrayList<Integer>();
			ArrayList<Integer> rightSideRule = new ArrayList<Integer>();
			HashMap<String,Integer> leftSideCounter = new HashMap<String,Integer>();
			HashMap<String,Integer> rightSideCounter = new HashMap<String,Integer>();
			Integer token = 0;
			String in = ""; String out = "";
			
			while(itSetPlaces.hasNext()) { // loop for a rule
				Place p = itSetPlaces.next();
				PlaceObject po = placeObjects.get(p);
				
				// involve marking in initial places
				if(initialPlaces.containsKey(p)) {
					token = token + initialPlaces.get(p);
					in = in + " token"+ token + " ";
				}
				
				ArrayList<String> transitionIn = po.transitionInEdge;
				ArrayList<String> transitionOut = po.transitionOutEdge;
				
				for(int j = 0; j < transitionIn.size(); j++) {
					String label = transitionIn.get(j);
					Integer counter = 0;
					if(leftSideCounter.containsKey(label)) {
						counter = 1 + leftSideCounter.get(label);
					} else {
						counter = 1;
					}
					leftSideCounter.put(label, counter);
				}
				for(int j = 0; j < transitionOut.size(); j++) {
					String label = transitionOut.get(j);
					Integer counter = 0;
					if(rightSideCounter.containsKey(label)) {
						counter = 1 + rightSideCounter.get(label);
					} else {
						counter = 1;
					}
					rightSideCounter.put(label, counter);
				}
			} // end of rule loop
			
			// create the left side of a rule
			for(String label : leftSideCounter.keySet()) {
				int count; 
				if(duplicateLabels.contains(label)) {		
					if(leftSideCounter.get(label) % transitionNodes.get(label).size() == 0) {
						count = leftSideCounter.get(label) / transitionNodes.get(label).size();		
					} else {
						count = leftSideCounter.get(label) / transitionNodes.get(label).size() + 1;
					}
				} else {
					count = leftSideCounter.get(label);
				}
				
				Integer intLabel;
				if(rule.eventClassToInt.containsKey(label)) {
					intLabel = rule.eventClassToInt.get(label);
				} else {
					intLabel = globalIndex;
					rule.eventClassToInt.put(label, intLabel);
					globalIndex++;
				}
				
				for(int j = 1; j <= count; j++) {
					leftSideRule.add(intLabel);
					in = in + label + " ";
				}
			}
			
			// create the right side of a rule
			for(String label : rightSideCounter.keySet()) {
				int count; 
				if(duplicateLabels.contains(label)) {
					if(rightSideCounter.get(label) % transitionNodes.get(label).size() == 0) {
						count = rightSideCounter.get(label) / transitionNodes.get(label).size();		
					} else {
						count = rightSideCounter.get(label) / transitionNodes.get(label).size() - 1;
					}
				} else {
					count = rightSideCounter.get(label);
				}
				
				Integer intLabel;
				if(rule.eventClassToInt.containsKey(label)) {
					intLabel = rule.eventClassToInt.get(label);
				} else {
					intLabel = globalIndex;
					rule.eventClassToInt.put(label, intLabel);
					globalIndex++;
				}
				
				for(int j = 1; j <= count; j++) {
					rightSideRule.add(intLabel);
					out = out +  label + " ";
				}
			}
			
			// does not involve final places
			if(!rightSideRule.isEmpty()) {
				if(token == 0) {
					AbstractMap.SimpleEntry<ArrayList<Integer>,ArrayList<Integer>> pair = new AbstractMap.SimpleEntry<ArrayList<Integer>,ArrayList<Integer>>(leftSideRule, rightSideRule);
					rule.rules.put(indexRule,pair);
					
					// create a map between activity to rule associated with it
					for(int i = 0; i < rightSideRule.size(); i++) {
						Integer actInt = rightSideRule.get(i);
						ArrayList<Integer> indexesRules;
						if(rule.rulesOfaLabel.containsKey(actInt)) {
							indexesRules = rule.rulesOfaLabel.get(actInt);							
						} else {
							indexesRules = new ArrayList<Integer>();
						}
						indexesRules.add(indexRule);
						rule.rulesOfaLabel.put(actInt, indexesRules);
					}
					
					indexRule++;
				} else {
					AbstractMap.SimpleEntry<Integer,ArrayList<Integer>> pairToken = new AbstractMap.SimpleEntry<Integer,ArrayList<Integer>>(token, leftSideRule);
					AbstractMap.SimpleEntry<AbstractMap.SimpleEntry<Integer,ArrayList<Integer>>, ArrayList<Integer>> pair = new AbstractMap.SimpleEntry<AbstractMap.SimpleEntry<Integer,ArrayList<Integer>>, ArrayList<Integer>>(pairToken, rightSideRule);
					rule.rulesWithToken.put(indexRuleToken,pair);
					
					// create a map between activity to rule associated with it
					for(int i = 0; i < rightSideRule.size(); i++) {
						Integer actInt = rightSideRule.get(i);
						ArrayList<Integer> indexesRules;
						if(rule.rulesTokenOfaLabel.containsKey(actInt)) {
							indexesRules = rule.rulesTokenOfaLabel.get(actInt);							
						} else {
							indexesRules = new ArrayList<Integer>();
						}
						indexesRules.add(indexRuleToken);
						rule.rulesTokenOfaLabel.put(actInt, indexesRules);
					}
					
					indexRuleToken++;
				}
				
				System.out.println("rule comb : " +  in + " >= " + out);
			}			
		}
	}
	
	/**
	 * combine places which made up a transition label, i.e.,
	 * places are combined because it shares duplicate labels. 
	 * e.g. A is made up from A1 and A2. then the places related to A1 and A2 must be combined properly.
	 * 
	 * @param places, the places related to a transition
	 * @param label, the transition label
	 * @param data, temporary array to store the place combination
	 * @param index
	 * @param length
	 * 
	 */
	private void combinePlaces(HashMap<Transition,ArrayList<Place>> places, String label, String ruleType, Place[] data, int index, int length) {
	    	
		// base case
		if (index == length) {
			ArrayList<ArrayList<Place>> list;
			if(ruleTypeAndCombinedPlaces.containsKey(ruleType)) {
				list = ruleTypeAndCombinedPlaces.get(ruleType);
			} else {
				list = new ArrayList<ArrayList<Place>>();
			}
			
			ArrayList<Place> newList = new ArrayList<Place>();
			for(int i = 0; i < data.length; i++) {
				newList.add(data[i]);
			}
			list.add(newList);
			ruleTypeAndCombinedPlaces.put(ruleType, list); // add place combination to ruleTypeAndCombinedPlaces, e.g. add [p1,p3] to the list of A_backward
			return;
		}
		
		int size = 0;
		if(places.get(transitionNodes.get(label).get(index)) != null) { // the transition does not have a place
			size = places.get(transitionNodes.get(label).get(index)).size();
		}
		
		for (int i = 0; i < size; i++){
			data[index] = places.get(transitionNodes.get(label).get(index)).get(i);			
			combinePlaces(places, label, ruleType, data, index+1, length); // recursive call
		}
			
	}
	
	/**
	 * create tree rule with backward direction
	 */
	private void createTreeRuleBackward() {
		
		// for each duplicate labels
		Iterator<String> itDuplicate = duplicateLabels.iterator();
		while(itDuplicate.hasNext()) {
			String root = itDuplicate.next();
			
			ArrayList<Transition> transitionNodeofRoot = new ArrayList<Transition>();	
			transitionNodeofRoot = transitionNodes.get(root);
			
			// for each transition of the root
			for(int i = 0; i < transitionNodeofRoot.size(); i++) {
				
				Transition transition = transitionNodeofRoot.get(i);
				
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdgesT = petrinet.getInEdges(transition);
				Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itInEdgesT = inEdgesT.iterator();
				
				// for each in-edge, take the source place
				while(itInEdgesT.hasNext()) {
					
					Place place = (Place) itInEdgesT.next().getSource();
					PlaceObject placeObject = new PlaceObject(place);
										
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdgesP = petrinet.getInEdges(place);
					Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itInEdgesP = inEdgesP.iterator();
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdgesP = petrinet.getOutEdges(place);
					Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itOutEdgesP = outEdgesP.iterator();
				
					// for each in-edge, take the source transition
					while(itInEdgesP.hasNext()) {
						
						Transition transitionIn = (Transition) itInEdgesP.next().getSource();
						
						// put transitionIn to placeObject
						placeObject.transitionInEdge.add(transitionIn.getLabel());
						// link the label duplication to the rule
						if(duplicateLabels.contains(transitionIn.getLabel())) {
							placeObject.links.add(transitionIn.getLabel() + "_forward");
						} 							
					}
					
					// for each out-edge, take the target transition
					while(itOutEdgesP.hasNext()){
						
						Transition transitionOut = (Transition) itOutEdgesP.next().getTarget();
						
						// put transitionOut to placeObject
						placeObject.transitionOutEdge.add(transitionOut.getLabel());
						// link the label duplication to the rule
						if(duplicateLabels.contains(transitionOut.getLabel())) {
							placeObject.links.add(transitionOut.getLabel() + "_backward");
						} 
						
					}
					
					// put place in 'placeBackward'
					ArrayList<Place> pList;
					if(placeBackward.containsKey(transition)) {
						pList = placeBackward.get(transition);
					} else {
						pList = new ArrayList<Place>();
					}
					pList.add(place);
					placeBackward.put(transition, pList);
					
					// put placeObject
					placeObjects.put(place, placeObject);
					
				}				
			}
		}
	}
	
	/**
	 * create tree rule with forward direction
	 */
	private void createTreeRuleForward() {
		// for each duplicate labels
		Iterator<String> itDuplicate = duplicateLabels.iterator();
		while(itDuplicate.hasNext()) {
			String root = itDuplicate.next();
			
			ArrayList<Transition> transitionNodeofRoot = new ArrayList<Transition>();	
			transitionNodeofRoot = transitionNodes.get(root);
			
			// for each transition of the root
			for(int i = 0; i < transitionNodeofRoot.size(); i++) {
				
				Transition transition = transitionNodeofRoot.get(i);
				
				Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdgesT = petrinet.getOutEdges(transition);
				Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itOutEdgesT = outEdgesT.iterator();
				
				// for each in-edge, take the source place
				while(itOutEdgesT.hasNext()) {
					
					Place place = (Place) itOutEdgesT.next().getTarget();
					PlaceObject placeObject = new PlaceObject(place);
					
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdgesP = petrinet.getOutEdges(place);
					Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itOutEdgesP = outEdgesP.iterator();
					Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdgesP = petrinet.getInEdges(place);
					Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> itInEdgesP = inEdgesP.iterator();
					
					// for each out-edge, take the target transition
					while(itOutEdgesP.hasNext()) {
						
						Transition transitionOut = (Transition) itOutEdgesP.next().getTarget();
						
						// put transitionOut to placeObject
						placeObject.transitionOutEdge.add(transitionOut.getLabel());
						// link the label duplication to the rule
						if(duplicateLabels.contains(transitionOut.getLabel())) {
							placeObject.links.add(transitionOut.getLabel() + "_backward");
						} 
							
					}
					
					// for each in-edge, take the source transition
					while(itInEdgesP.hasNext()){
						
						Transition transitionIn = (Transition) itInEdgesP.next().getSource();
						
						// put transitionIn to placeObject
						placeObject.transitionInEdge.add(transitionIn.getLabel());						
						// link the label duplication to the rule
						if(duplicateLabels.contains(transitionIn.getLabel())) {
							placeObject.links.add(transitionIn.getLabel() + "_forward");
						} 
						
					}
					
					// put place in 'placeForward'
					ArrayList<Place> pList;
					if(placeForward.containsKey(transition)) {
						pList = placeForward.get(transition);
					} else {
						pList = new ArrayList<Place>();
					}
					pList.add(place);
					placeForward.put(transition, pList);
					
					// put placeObject
					placeObjects.put(place, placeObject);
										
				}
			}
		}
	}
	
	public Rule getRule() {
		System.out.println("total place-based rules generated: " + (indexRule + indexRuleToken));
		return this.rule;
	}	
}
