package org.processmining.indatabasepreprocessing.dialogs;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.ListSelectionModel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.util.ui.widgets.ProMList;
import org.processmining.framework.util.ui.widgets.ProMPasswordField;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.indatabasepreprocessing.algorithms.BehaviouralProfileProduction;
import org.processmining.indatabasepreprocessing.algorithms.RuleProduction;
import org.processmining.indatabasepreprocessing.parameters.DBConnectionParameter;
import org.processmining.indatabasepreprocessing.parameters.MiningAlgorithm;
import org.processmining.indatabasepreprocessing.parameters.Rule;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;

import com.fluxicon.slickerbox.factory.SlickerFactory;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstants;

public class DBConnectionDialog {
	private UIPluginContext context;
	private DBConnectionParameter parameter;
	private int databaseStep;
	private int pmAlgorithmStep;
	private int incrementalStep;
	private int currentStep;
	private int nofSteps;
	private SetupStep[] steps;
	private boolean createRule;
	private boolean createBehaviouralProfile;
	private Petrinet net;
	private Marking marking;
	
	public DBConnectionDialog(UIPluginContext context) {
		this.context = context;
	}
	
	public DBConnectionDialog(UIPluginContext context, Petrinet net) {
		this.context = context;
		this.net = net;
		checkMarking();
	}
		
	public DBConnectionParameter setup() {
		parameter = new DBConnectionParameter();
		createRule = false;
		InteractionResult result = InteractionResult.NEXT;
		
		nofSteps = 0;
		databaseStep = nofSteps++;
		pmAlgorithmStep = nofSteps++;
		incrementalStep = nofSteps++;
		
		steps = new SetupStep[nofSteps];
		steps[databaseStep] = new DatabaseStep();
		steps[pmAlgorithmStep] = new PMAlgorithmStep();
		steps[incrementalStep] = new IncrementalStep();
		
		currentStep = databaseStep;
		
		while (true) {
			if (currentStep < 0) {
				currentStep = 0;
			}
			if (currentStep >= nofSteps) {
				currentStep = nofSteps - 1;
			}
			
			result = context.showWizard("DB Connection Manager", currentStep == 0, currentStep == nofSteps - 1, steps[currentStep]);
			switch (result) {
				case NEXT :
					steps[currentStep].readSettings();
					go(1);
					break;
				case PREV :
					go(-1);
					break;
				case FINISHED :
					steps[currentStep].readSettings();
					return parameter; 
				default :
					return null;
			}
		}
	}
	
	private int go(int direction) {
		currentStep += direction;
		if (steps[currentStep].precondition()) {
			return currentStep;
		} else {
			currentStep -= direction;
			return currentStep;
		}
	}
	
	private abstract class SetupStep extends JPanel {
		private static final long serialVersionUID = 8889689637540127545L;
		public abstract boolean precondition();
		public abstract void readSettings();
	}
	
	private class DatabaseStep extends SetupStep {
		private static final long serialVersionUID = 8991567528446950633L;
		private ProMTextField server;
		private ProMTextField username;
		private ProMPasswordField password;	
		private ProMTextField databaseName;
		private ProMTextField logTable;

		public DatabaseStep() {
			initComponents();
		}
		
		private void initComponents() {
			double size[][] = { { 200, 5, 550 },
					{ 50, 20, 30, 30, 30, 30, 30, TableLayoutConstants.FILL } };
			TableLayout layout = new TableLayout(size);
			setLayout(layout);
			
			server = new ProMTextField(parameter.getServer(), "Enter your database server name");
			username = new ProMTextField(parameter.getUsername(), "Enter your database username");
			password = new ProMPasswordField(parameter.getPassword());
			databaseName = new ProMTextField(parameter.getDatabaseName(), "Enter your database name");
			logTable = new ProMTextField(parameter.getLogTable(), "Enter your log table");
		
			add(SlickerFactory.instance().createLabel("<html><h2>Configure database connection</h2>"), "0,0,2,0");
			add(SlickerFactory.instance().createLabel("<html><h3>Database server name:</h3>"), "0,2");
			add(server, "2,2");
			add(SlickerFactory.instance().createLabel("<html><h3>Username:</h3>"), "0,3");
			add(username, "2,3");
			add(SlickerFactory.instance().createLabel("<html><h3>Password:</h3>"), "0,4");
			add(password,  "2,4");
			add(SlickerFactory.instance().createLabel("<html><h3>Database name:</h3>"), "0,5");
			add(databaseName,  "2,5");
			add(SlickerFactory.instance().createLabel("<html><h3>Log table:</h3>"), "0,6");
			add(logTable,  "2,6");
		}
		
		public boolean precondition() {
			return true;
		}

		public void readSettings() {
			parameter.setServer(server.getText());
			parameter.setUsername(username.getText());
			parameter.setPassword(password.getText());
			parameter.setDatabaseName(databaseName.getText());
			parameter.setLogTable(logTable.getText());	
			
			// only for experiment
			parameter.setDatabaseName("test");
			//parameter.setLogTable("log_abc");
			parameter.setPassword("");
			parameter.setUsername("sa");
			parameter.setServer("localhost");
		}

	}
	

	private class PMAlgorithmStep extends SetupStep {
		private static final long serialVersionUID = 7509826099209147073L;
		private ProMList<String> algoList;
		private ProMList<String> chosenAlgo;
		private DefaultListModel<String> algoListModel;
		private DefaultListModel<String> chosenAlgoModel;

		public PMAlgorithmStep() {
			initComponents();
		}
		
		private void initComponents() {
			algoListModel = new DefaultListModel<String>();
			algoListModel.addElement("Inductive Miner");
			algoListModel.addElement("Declare MINERful");
			algoListModel.addElement("Flexible Heuristics Miner");
			algoListModel.addElement("Data Aware Heuristics Miner");
			algoListModel.addElement("Absolute Handover of Work");
			algoListModel.addElement("Boolean Handover of Work");
			algoListModel.addElement("Absolute Causal Handover of Work");			
			algoListModel.addElement("Boolean Causal Handover of Work");			
			algoListModel.addElement("Alignment with Place-based Rules");
			algoListModel.addElement("Alignment with Behavioural-based Rules");
			//algoListModel.addElement("Behavioural profile");
			algoListModel.addElement("Performance Analysis");
			algoListModel.addElement("Performance Analysis No Checking");
			
			algoList = new ProMList<String>("List of Process Mining Algorithms", algoListModel);
			algoList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			
			chosenAlgoModel = new DefaultListModel<String>();
			chosenAlgo = new ProMList<String>("Chosen Process Mining Algorithms", chosenAlgoModel);
			
			JButton addAlgoButton = SlickerFactory.instance().createButton("Add Algorithm >>");
			addAlgoButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (!algoList.getSelectedValuesList().isEmpty()) {
						if ((chosenAlgoModel.size() == 1) && (chosenAlgoModel.elementAt(0).equals("EMPTY"))) {
							chosenAlgoModel.removeAllElements();
						}
						chosenAlgoModel.addElement(algoList.getSelectedValuesList().get(0));
					}
				}
			});
			
			JButton removeAlgoButton = SlickerFactory.instance().createButton("<< Remove Algorithm");
			removeAlgoButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					for (Object obj : chosenAlgo.getSelectedValuesList()) {
						chosenAlgoModel.removeElement(obj);
					}
				}
			});
			
			double[][] size = new double[][] { { 250, 10, 200, 10, 250 }, { 40, TableLayout.FILL, 30, 5, 30, TableLayout.FILL } };
			TableLayout layout = new TableLayout(size);
			setLayout(layout);
			add(SlickerFactory.instance().createLabel("<html><h2>Configure process mining algorithm</h2>"), "0,0,2,0");
			add(algoList, "0,1,0,5");
			add(addAlgoButton, "2,2");
			add(removeAlgoButton, "2,4");
			add(chosenAlgo, "4,1,4,5");
		}

		public boolean precondition() {			
			Connection conn = null;
			try {
				Class.forName("org.h2.Driver");
				conn = DriverManager.getConnection("jdbc:h2:tcp://"+parameter.getServer()+"/~/"+parameter.getDatabaseName()+";LOG=0;CACHE_SIZE=2048000000;LOCK_MODE=0;UNDO_LOG=0;REFERENTIAL_INTEGRITY=FALSE;CACHE_TYPE=TQ", 
						parameter.getUsername(), parameter.getPassword());
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} catch (SQLException e) {
				e.printStackTrace();
			}		
			
			if(conn == null) {
				JOptionPane.showConfirmDialog(
						new JPanel(),
						"Database connection cannot be made. Please check the configuration again.",
						"No Database Connection", JOptionPane.PLAIN_MESSAGE, JOptionPane.INFORMATION_MESSAGE);
				return false;
			} else {				
				return true;
			}
			
		}

		public void readSettings() {
			MiningAlgorithm[] miningAlgo = new MiningAlgorithm[chosenAlgoModel.size()];
			for(int i = 0; i < chosenAlgoModel.size(); i++) {
				switch(chosenAlgoModel.get(i)) {
					case "Inductive Miner" : miningAlgo[i] = MiningAlgorithm.INDUCTIVE_MINER; break;
					case "Declare MINERful" : miningAlgo[i] = MiningAlgorithm.DECLARE_MINERFUL; break;
					case "Flexible Heuristics Miner" : miningAlgo[i] = MiningAlgorithm.FLEXIBLE_HEURISTIC_MINER; break;
					case "Data Aware Heuristics Miner" : miningAlgo[i] = MiningAlgorithm.DATA_AWARE_HEURISTIC_MINER; break;
					case "Absolute Causal Handover of Work" : miningAlgo[i] = MiningAlgorithm.ABSOLUTE_CAUSAL_HANDOVER_OF_WORK; break;
					case "Absolute Handover of Work" : miningAlgo[i] = MiningAlgorithm.ABSOLUTE_HANDOVER_OF_WORK; break;
					case "Boolean Causal Handover of Work" : miningAlgo[i] = MiningAlgorithm.BOOLEAN_CAUSAL_HANDOVER_OF_WORK; break;
					case "Boolean Handover of Work" : miningAlgo[i] = MiningAlgorithm.BOOLEAN_HANDOVER_OF_WORK; break;
					case "Alignment with Place-based Rules" : miningAlgo[i] = MiningAlgorithm.ALIGNMENT_RULE_CHECKING; createRule = true; break;
					case "Alignment with Behavioural-based Rules" : miningAlgo[i] = MiningAlgorithm.ALIGNMENT_BEHAVIOURAL; createBehaviouralProfile = true; break;
					case "Performance Analysis" : miningAlgo[i] = MiningAlgorithm.PERFORMANCE_ANALYSIS; createRule = true; break;
					case "Performance Analysis No Checking" : miningAlgo[i] = MiningAlgorithm.PERFORMANCE_ANALYSIS_NOCHECK; break;
				}
			}
			parameter.setPMAlgorithms(miningAlgo);
		}
		
	}
	
	private class IncrementalStep extends SetupStep {
		private static final long serialVersionUID = 3816661754749839478L;
		private JRadioButton yesInvolve;
		private JRadioButton noInvolve;

		public IncrementalStep() {
			initComponents();
		}
		
		private void initComponents() {
			double size[][] = { { 54, TableLayoutConstants.FILL },
					{ 50, 30, 20, 20, 10, 30, 20, 20, 20, 20, 20, 20, TableLayoutConstants.FILL } };
			TableLayout layout = new TableLayout(size);
			setLayout(layout);
						
			yesInvolve = SlickerFactory.instance().createRadioButton("Yes");
			noInvolve = SlickerFactory.instance().createRadioButton("No");
			ButtonGroup involveGroup = new ButtonGroup();
			involveGroup.add(yesInvolve);
			involveGroup.add(noInvolve);
			yesInvolve.setSelected(true);
			
			add(SlickerFactory.instance().createLabel("<html><h2>Configure parameter for incremental discovery/conformance checking </h2>"), "0,0,1,0");
			add(SlickerFactory.instance().createLabel("<html><h3>Involve existing data:</h3>"), "0,5,1,5");
			add(yesInvolve, "0,6,1,6");
			add(noInvolve, "0,7,1,7");
		}

		public boolean precondition() {			
			
			if(parameter.getPMAlgorithms().length == 0) {
				JOptionPane.showConfirmDialog(
						new JPanel(),
						"No algorithm is chosen. Please choose one or more.",
						"No Process Mining Algorithm", JOptionPane.PLAIN_MESSAGE, JOptionPane.INFORMATION_MESSAGE);
				
				return false;
			} else {
				return true;
			}
			
		}

		public void readSettings() {		
			
			if(yesInvolve.isSelected()) {
				parameter.setPreviousData(true);
			} else {
				parameter.setPreviousData(false);
			}
			
			//if alignment/performance analysis is chosen, create rule
			if(createRule) {
				ruleProduction();
			}
			
			// if alignment with behavioural profile is chosen, create profile
			if(createBehaviouralProfile) {
				behaviouralProduction();
			}
		}
		
		private void ruleProduction() {
			// check if petri net exists
			if(net == null) {
				JOptionPane.showMessageDialog(new JPanel(),
						"No rule checking can be performed as the net does not exist.");
				context.getFutureResult(0).cancel(true);
				return;
			}
			
			RuleProduction rp = new RuleProduction(net, marking);		
			Rule rule = rp.getRule();
			String ruleStr = rule.ruleToString() + "\n" + rule.ruleWithTokenToString() + "\n" + rule.eventClassIntToString() + "\n" + rule.ruleOfALabelToString() + "\n" + rule.ruleTokenOfALabelToString();
			
			parameter.setRule(ruleStr);
			parameter.setEventClassToInt(rule.eventClassToInt);
						
		}
		
		private void behaviouralProduction() {
			// check if reachability graph exists
			if(net == null) {
				JOptionPane.showMessageDialog(new JPanel(),
						"No behavioural profile can be performed as the net does not exist.");
				context.getFutureResult(0).cancel(true);
				return;
			}
			
			BehaviouralProfileProduction bp = null;
			try {
				bp = new BehaviouralProfileProduction(context, net, marking);
			} catch (Exception e) {
				e.printStackTrace();
			}
			bp.createBehaviouralProfile();
			String behavStr = bp.toString();
			
			parameter.setBehaviouralProfile(behavStr);
			
		}
		
	}
	
	private void checkMarking() {
		try {
			InitialMarkingConnection con = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, net);
			this.marking = (Marking) con.getObjectWithRole(InitialMarkingConnection.MARKING);
			
		} catch (ConnectionCannotBeObtained conProb) {
			JOptionPane.showMessageDialog(new JPanel(),
					"No rule checking can be performed as the net does not have any corresponding initial marking");
			context.getFutureResult(0).cancel(true);
			return;
		}
	}
}
